from votekit.pref_profile import ScoreProfile, CleanedScoreProfile, ProfileError
from typing import Callable
import pandas as pd
from typing import Union
from functools import partial
import numpy as np


def _iterate_and_clean_score_tuples(
    profile: ScoreProfile,
    clean_score_func: Callable[[tuple], tuple],
) -> tuple[pd.DataFrame, set[int], set[int], set[int], set[int]]:
    """
    Clean the rows of the df according to the cleaning rule. Note that this function
    only allows you to edit the score tuple, not any other part of the ballot.

    Args:
        profile (ScoreProfile): The original profile of ballots.
        clean_score_func (Callable[[tuple], tuple]): Function that
            takes the score portion of a row of the profile df and returns an altered score tuple.

    Returns:
        tuple[pd.DataFrame, set[int], set[int], set[int], set[int]]: cleaned_df, unaltr_idxs,
            nonempty_altr_idxs, no_wt_altr_idxs, no_scores_altr_idxs
    """
    cleaned_df = profile.df.copy()
    candidate_cols = [c for c in cleaned_df.columns if c not in ["Weight", "Voter Set"]]

    orig_rows = [
        tuple(vals)
        for vals in cleaned_df[candidate_cols].itertuples(index=False, name=None)
    ]
    cleaned_rows = [clean_score_func(row) for row in orig_rows]

    cleaned_df[candidate_cols] = pd.DataFrame(cleaned_rows, index=cleaned_df.index)
    idxs = cleaned_df.index

    unaltr_idxs = {
        idx
        for idx, (o, c) in zip(idxs, zip(orig_rows, cleaned_rows))
        if np.array_equal(o, c, equal_nan=True)
    }
    no_scores_altr_idxs = {
        idx
        for idx, score_tuple in zip(idxs, cleaned_rows)
        if all(score == 0 or pd.isna(score) for score in score_tuple)
    }
    nonempty_altr_idxs = set(idxs) - unaltr_idxs - no_scores_altr_idxs
    no_wt_altr_idxs: set[int] = set()

    return (
        cleaned_df,
        unaltr_idxs,
        nonempty_altr_idxs,
        no_wt_altr_idxs,
        no_scores_altr_idxs,
    )


def clean_score_profile(
    profile: ScoreProfile,
    clean_score_func: Callable[[tuple], tuple],
    remove_empty_ballots: bool = True,
    remove_zero_weight_ballots: bool = True,
    retain_original_candidate_list: bool = True,
) -> CleanedScoreProfile:
    """
    Allows user-defined cleaning rules for ScoreProfile. Cleaning function can only
    be applied to the score section of the dataframe, not the weight or voter set.


    Args:
        profile (ScoreProfile): A ScoreProfile to clean.
        clean_score_func (Callable[[tuple], tuple]): Function that
            takes the score portion of a row of the profile df and returns an altered score.
        remove_empty_ballots (bool, optional): Whether or not to remove ballots that have no
            no scores as a result of the cleaning. Defaults to True.
        remove_zero_weight_ballots (bool, optional): Whether or not to remove ballots that have no
            weight as a result of the cleaning. Defaults to True.
        retain_original_candidate_list (bool, optional): Whether or not to use the candidate list
            from the original profile in the new profile. If False, uses only candidates who receive
            votes. Defaults to True.

    Returns:
        CleanedScoreProfile: A cleaned ``ScoreProfile``.

    Raises:
        ProfileError: Profile must only contain ranked ballots.
    """
    if not isinstance(profile, ScoreProfile):
        raise ProfileError("Profile must be a ScoreProfile.")

    (
        cleaned_df,
        unaltr_idxs,
        nonempty_altr_idxs,
        no_wt_altr_idxs,
        no_scores_altr_idxs,
    ) = _iterate_and_clean_score_tuples(profile, clean_score_func)

    if remove_empty_ballots:
        candidate_cols = [
            c for c in cleaned_df.columns if c not in ["Weight", "Voter Set"]
        ]
        mask = (
            cleaned_df[candidate_cols]
            .map(lambda score: score == 0 or pd.isna(score))  # type: ignore[operator]
            .all(axis=1)
        )

        cleaned_df = cleaned_df[~mask]

    if remove_zero_weight_ballots:
        cleaned_df = cleaned_df[cleaned_df["Weight"] > 0]

    return CleanedScoreProfile(
        df=cleaned_df,
        candidates=(profile.candidates if retain_original_candidate_list else tuple()),
        no_wt_altr_idxs=no_wt_altr_idxs,
        no_scores_altr_idxs=no_scores_altr_idxs,
        nonempty_altr_idxs=nonempty_altr_idxs,
        unaltr_idxs=unaltr_idxs,
        parent_profile=profile,
        df_index_column=list(cleaned_df.index),
    )


def remove_cand_from_score_tuple(
    removed_idxs: list[int],
    score_tup: tuple[float, ...],
) -> tuple[float, ...]:
    """
    Removes specified candidate(s) from score tuple.

    Args:
        removed_idxs (list[int]): Indices of candidates to be removed.
        ranking_tup (tuple): Ranking to remove candidates from.

    Returns:
        tuple: Ranking with candidate(s) removed.
    """
    return tuple(
        s if i not in removed_idxs else np.nan for i, s in enumerate(score_tup)
    )


def remove_cand_score_profile(
    removed: Union[str, list],
    profile: ScoreProfile,
    remove_empty_ballots: bool = True,
    remove_zero_weight_ballots: bool = True,
    retain_original_candidate_list: bool = False,
) -> CleanedScoreProfile:
    """
    Given a score profile, remove the given candidate(s) from the ballots.

    Wrapper for clean_score_profile that does some extra processing to ensure the candidate list
    is handled correctly.

    Not as fast as removing the candidate columns directly, but tracks more information
    about which ballots were adjusted.

    Args:
        removed (Union[str, list]): Candidate or list of candidates to be removed.
        profile (ScoreProfile): Profile to remove candidates from.
        remove_empty_ballots (bool, optional): Whether or not to remove ballots that have no
            ranking or scores as a result of cleaning. Defaults to True.
        remove_zero_weight_ballots (bool, optional): Whether or not to remove ballots that have no
            weight as a result of cleaning. Defaults to True.
        retain_original_candidate_list (bool, optional): Whether or not to use the candidate list
            from the orginal profile in the new profile. If False, takes the original candidate
            list and removes the candidate(s) given in ``removed``, but preserves all others.
            Defaults to False.

    Returns:
        CleanedScoreProfile: A cleaned ``ScoreProfile``.

    Raises:
        ProfileError: Profile must only contain score ballots.
    """
    if isinstance(removed, str):
        removed = [removed]
    removed_idxs = [i for i, c in enumerate(profile.df.columns) if c in removed]

    cleaned_profile = clean_score_profile(
        profile,
        partial(remove_cand_from_score_tuple, removed_idxs),
        remove_empty_ballots,
        remove_zero_weight_ballots,
        retain_original_candidate_list=True,
    )

    new_candidates = (
        [c for c in profile.candidates if c not in removed]
        if not retain_original_candidate_list
        else list(profile.candidates)
    )
    new_df = (
        cleaned_profile.df[new_candidates + ["Voter Set", "Weight"]]
        if not retain_original_candidate_list
        else cleaned_profile.df
    )

    return CleanedScoreProfile(
        df=new_df,
        candidates=new_candidates,
        parent_profile=cleaned_profile.parent_profile,
        df_index_column=cleaned_profile.df_index_column,
        no_wt_altr_idxs=cleaned_profile.no_wt_altr_idxs,
        no_scores_altr_idxs=cleaned_profile.no_scores_altr_idxs,
        nonempty_altr_idxs=cleaned_profile.nonempty_altr_idxs,
        unaltr_idxs=cleaned_profile.unaltr_idxs,
    )



from votekit.ballot import RankBallot
from votekit.pref_profile import RankProfile
from votekit.matrices._utils import _convert_dict_to_matrix
import numpy as np
from typing import Union


def comention(cands: Union[str, list[str]], ballot: RankBallot):
    """
    Takes cands and returns true if they all appear on the ballot in the ranking.

    Args:
      cands (Union[str, list[str]]): Candidate name or list of candidate names.
      ballot (RankBallot): RankBallot.

    Returns:
      bool: True if all candidates appear in ballot.
    """
    all_cands: set[str] = set()

    if ballot.ranking:
        all_cands = all_cands.union(c for s in ballot.ranking for c in s)

    if isinstance(cands, str):
        cands = [cands]

    return set(cands).issubset(all_cands)


def comention_above(i: str, j: str, ballot: RankBallot) -> bool:
    """
    Takes candidates i,j and returns True if i >= j in the ranking.
    Requires that the ballot has a ranking.


    Args:
      i (str): Candidate name.
      j (str): Candidate name.
      ballot (RankBallot): RankBallot.

    Returns:
      bool: True if both i and j appear in ballot and i >= j.
    """
    if not isinstance(ballot, RankBallot):
        raise TypeError("Ballot must be of type RankBallot.")
    if ballot.ranking is None:
        raise TypeError(f"RankBallot must have a ranking: {ballot}")
    i_index, j_index = (-1, -1)

    for rank, s in enumerate(ballot.ranking):
        if i in s:
            i_index = rank
        if j in s:
            j_index = rank

    return (i_index >= 0 and j_index >= 0) and (i_index <= j_index)


def comentions_matrix(
    pref_profile: RankProfile, candidates: list[str], symmetric: bool = False
) -> np.ndarray:
    """
    Takes a preference profile and converts to a matrix
    where the i,j entry shows the number of times candidates i,j were mentioned on the same
    ballot with i >= j. There is an option to make it symmetric so that the i,j entry is just
    the number of times candidates i and j were mentioned on the same ballot.

    Args:
      pref_profile (RankProfile): Profile.
      candidates (list[str]): List of candidates to use. Indexing of this list matches indexing of
        output array.
      symmetric (bool, optional): Whether or not to make the matrix symmetric. Defaults to False
        in which case the i,j entry is comentions where i >= j. True means the i,j entry is
        comentions of i,j.

    Returns:
      np.ndarray: Numpy array of comentions.
    """
    comentions_matrix = {c: {c: 0.0 for c in candidates} for c in candidates}

    for i in candidates:
        for j in candidates:
            for ballot in pref_profile.ballots:

                if symmetric:
                    if comention([i, j], ballot):
                        comentions_matrix[i][j] += ballot.weight
                else:
                    if comention_above(i, j, ballot):
                        comentions_matrix[i][j] += ballot.weight

    return _convert_dict_to_matrix(comentions_matrix)



from .pref_profile import PreferenceProfile, RankProfile, ScoreProfile, ProfileError
from .utils import (
    rank_profile_to_ballot_dict,
    score_profile_to_ballot_dict,
    rank_profile_to_ranking_dict,
    score_profile_to_scores_dict,
    profile_df_head,
    profile_df_tail,
    convert_row_to_rank_ballot,
    convert_rank_profile_to_score_profile_via_score_vector,
)
from .cleaned_pref_profile import CleanedRankProfile, CleanedScoreProfile

__all__ = [
    "PreferenceProfile",
    "RankProfile",
    "ScoreProfile",
    "ProfileError",
    "rank_profile_to_ballot_dict",
    "score_profile_to_ballot_dict",
    "rank_profile_to_ranking_dict",
    "score_profile_to_scores_dict",
    "profile_df_head",
    "profile_df_tail",
    "CleanedRankProfile",
    "CleanedScoreProfile",
    "convert_row_to_rank_ballot",
    "convert_rank_profile_to_score_profile_via_score_vector",
]



from __future__ import annotations
import csv
import os
import urllib.request
import pandas as pd
from votekit.ballot import Ballot, ScoreBallot, RankBallot
from votekit.pref_profile.utils import (
    convert_row_to_rank_ballot,
    convert_row_to_score_ballot,
)
from votekit.pref_profile.csv_utils import (
    _validate_rank_csv_format,
    _parse_profile_data_from_rank_csv,
    _parse_ballot_from_rank_csv,
    _validate_score_csv_format,
    _parse_profile_data_from_score_csv,
    _parse_ballot_from_score_csv,
)
import numpy as np
from typing import Optional, Tuple, Sequence, Union
import warnings
import pickle
from functools import cached_property
from pathlib import Path
from os import PathLike


class ProfileError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class PreferenceProfile:
    """
    PreferenceProfile class, contains ballots and candidates for a given election.
    This is a frozen class, so you need to create a new PreferenceProfile any time
    you want to edit the ballots, candidates, etc.

    Args:
        ballots (Sequence[Ballot], optional): Tuple of ``Ballot`` objects. Defaults to empty tuple.
        candidates (tuple[str], optional): Tuple of candidate strings. Defaults to empty tuple.
            If empty, computes this from any candidate listed on a ballot with positive weight.
        max_ranking_length (int, optional): The length of the longest allowable ballot, i.e., how
            many candidates are allowed to be ranked in an election. Defaults to longest observed
            ballot.
        contains_rankings (bool, optional): Whether or not the profile contains ballots with
            rankings. If no boolean is provided, then the appropriate boolean value will be
            interpreted from the input preference profile (i.e. if some ballot in the profile has
            a ranking, then this will be set  to `True`).
        contains_scores (bool, optional): Whether or not the profile contains ballots with
            scores. If no boolean is provided, then the appropriate boolean value will be
            interpreted from the input preference profile (i.e. if some ballot in the profile has
            a ranking, then this will be set  to `True`).

    Parameters:
        ballots (Sequence[Ballot]): Tuple of ``Ballot`` objects.
        candidates (tuple[str]): Tuple of candidate strings.
        max_ranking_length (int): The length of the longest allowable ballot, i.e., how
            many candidates are allowed to be ranked in an election.
        df (pandas.DataFrame): Data frame view of the ballots.
        candidates_cast (tuple[str]): Tuple of candidates who appear on any ballot with positive
            weight, either in the ranking or in the score dictionary.
        total_ballot_wt (float): Sum of ballot weights.
        num_ballots (int): Length of ballot list.
        contains_rankings (bool): Whether or not the profile contains ballots with
            rankings.
        contains_scores (bool): Whether or not the profile contains ballots with
            scores.

    Raises:
        ProfileError: a data frame and ballot list are passed to the init method.
        ProfileError: contains_rankings is set to False but a ballot contains a ranking.
        ProfileError: contains_rankings is set to True but no ballot contains a ranking.
        ProfileError: contains_scores is set to False but a ballot contains a score.
        ProfileError: contains_scores is set to True but no ballot contains a score.
        ProfileError: max_ranking_length is set but a ballot ranking excedes the length.
        ProfileError: a candidate is found on a ballot that is not listed on a provided
            candidate list.
        ProfileError: candidates must be unique.
        ProfileError: candidates must not have names matching ranking columns.

    """

    _is_frozen: bool = False

    def __new__(
        cls,
        *,
        ballots: Sequence[Ballot] = tuple(),
        candidates: Sequence[str] = tuple(),
        max_ranking_length: int = 0,
        df: pd.DataFrame = pd.DataFrame(),
        **kwargs,
    ):

        if not df.equals(pd.DataFrame()) and len(ballots) != 0:
            raise ProfileError(
                "Cannot pass a dataframe and a ballot list to profile init method. Must pick one."
            )

        if df.equals(pd.DataFrame()) and len(ballots) == 0:
            return super().__new__(cls)

        elif df.equals(pd.DataFrame()):
            if all(isinstance(b, RankBallot) for b in ballots):
                return super().__new__(RankProfile)

            elif all(isinstance(b, ScoreBallot) for b in ballots):
                return super().__new__(ScoreProfile)

            else:
                score_idxs = [
                    idx for idx, b in enumerate(ballots) if isinstance(b, ScoreBallot)
                ]
                rank_idxs = [
                    idx for idx, b in enumerate(ballots) if isinstance(b, RankBallot)
                ]

                raise ProfileError(
                    f"Profile cannot contain RankBallots and ScoreBallots. There are "
                    f"{len(score_idxs)} ScoreBallots and {len(rank_idxs)} RankBallots."
                )

        if any(c.startswith("Ranking_") for c in df.columns):
            return super().__new__(RankProfile)

        return super().__new__(ScoreProfile)

    def __init__(
        self,
        *,
        ballots: Sequence[Ballot] = tuple(),
        candidates_cast: Sequence[str] = tuple(),
        candidates: Sequence[str] = tuple(),
        max_ranking_length: Optional[int] = None,
        df: pd.DataFrame = pd.DataFrame(),
    ):
        self.candidates_cast = candidates_cast
        self.candidates = candidates
        self.max_ranking_length = max_ranking_length
        self.df = df

        self.total_ballot_wt = self._find_total_ballot_wt()
        self.num_ballots = self._find_num_ballots()
        self._validate_and_set_candidates()

        self._is_frozen = True

    def _find_num_ballots(self) -> int:
        """
        Compute and set the number of ballots.

        Returns:
            int: num ballots
        """
        return len(self.df)

    def _find_total_ballot_wt(self) -> float:
        """
        Compute and set the total ballot weight.

        Returns:
            float: total ballot weight.
        """
        total_weight = 0

        if not self.df.equals(pd.DataFrame()):
            total_weight = self.df["Weight"].sum()

        return total_weight

    def _validate_and_set_candidates(self) -> None:
        """
        Ensure that the candidate names are not equal to the ranking column names, that they are
        unique, and strips whitespace from candidates.

        Raises:
            ProfileError: Candidate names must not be the same as "Ranking_i".
            ProfileError: Candidate names must be unique.
        """
        for cand in self.candidates:
            if any(f"Ranking_{i}" == cand for i in range(len(self.candidates))):
                raise ProfileError(
                    (
                        f"{cand} is a name reserved for ranking columns, it cannot be used as a "
                        "candidate name."
                    )
                )

        if not len(set(self.candidates)) == len(self.candidates):
            raise ProfileError("All candidates must be unique.")

        if not set(self.candidates_cast).issubset(self.candidates):
            raise ProfileError(
                "Candidates cast are not a subset of candidates list. The following "
                " candidates are in candidates_cast but not candidates: "
                f"{set(self.candidates_cast)-set(self.candidates)}."
            )

        self.candidates = tuple([c.strip() for c in self.candidates])
        self.candidates_cast = tuple([c.strip() for c in self.candidates_cast])

    def __setattr__(self, name, value):
        if getattr(self, "_is_frozen", False):
            raise AttributeError(
                f"Cannot modify frozen instance: tried to set '{name}'"
            )
        super().__setattr__(name, value)

    def __delattr__(self, name):
        if getattr(self, "_is_frozen", False):
            raise AttributeError(
                f"Cannot delete attribute '{name}' from frozen instance"
            )
        super().__delattr__(name)

    def __eq__(self, other):
        if not isinstance(other, PreferenceProfile):
            return False
        if set(self.candidates) != set(other.candidates):
            return False
        if set(self.candidates_cast) != set(other.candidates_cast):
            return False
        if self.total_ballot_wt != other.total_ballot_wt:
            return False

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            pp_1 = self.group_ballots()
            pp_2 = other.group_ballots()
        for b in pp_1.ballots:
            if b not in pp_2.ballots:
                return False
        for b in pp_2.ballots:
            if b not in pp_1.ballots:
                return False
        return True

    def __str__(self) -> str:

        repr_str = "PrefenceProfile"
        repr_str += (
            f"Candidates: {self.candidates}\n"
            f"Candidates who received votes: {self.candidates_cast}\n"
            f"Total number of Ballot objects: {self.num_ballots}\n"
            f"Total weight of Ballot objects: {self.total_ballot_wt}\n"
        )

        return repr_str

    __repr__ = __str__

    def to_pickle(self, fpath: Union[str, PathLike, Path]):
        """
        Saves profile to pickle file.

        Args:
            fpath (Union[str, os.PathLike, pathlib.Path]): File path to save profile to.

        Raises:
            ValueError: File path must be provided.
        """
        if fpath == "":
            raise ValueError("File path must be provided.")
        with open(str(fpath), "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_pickle(cls, fpath: Union[str, PathLike, Path]) -> PreferenceProfile:
        """
        Reads profile from pickle file.

        Args:
            fpath (Union[str, os.PathLike, pathlib.Path]): File path to profile. Can be a url.
        """
        fpath = str(fpath)
        if not os.path.isfile(fpath):
            with urllib.request.urlopen(fpath) as response:
                data = pickle.loads(response.read())
        else:
            with open(str(fpath), "rb") as f:
                data = pickle.load(f)
        assert isinstance(data, PreferenceProfile)
        return data


class RankProfile(PreferenceProfile):

    def __init__(
        self,
        *,
        ballots: Sequence[RankBallot] = tuple(),
        candidates: Sequence[str] = tuple(),
        max_ranking_length: Optional[int] = None,
        df: pd.DataFrame = pd.DataFrame(),
    ):
        self.candidates = tuple(candidates)
        self.max_ranking_length = (
            0 if max_ranking_length is None else max_ranking_length
        )

        if df.equals(pd.DataFrame()):
            (
                self.df,
                self.candidates_cast,
            ) = self._init_from_rank_ballots(ballots)
            if self.candidates == tuple():
                self.candidates = self.candidates_cast

        else:
            self.df, self.candidates_cast = self._init_from_rank_df(df)

        self.max_ranking_length = self._find_max_ranking_length()

        super().__init__(
            candidates=self.candidates,
            candidates_cast=self.candidates_cast,
            df=self.df,
            max_ranking_length=self.max_ranking_length,
        )

    def __update_ballot_ranking_data(
        self,
        rank_ballot_data: dict[str, list],
        idx: int,
        rank_ballot: RankBallot,
        candidates_cast: list[str],
        num_ballots: int,
    ):
        """
        Update the ranking data from a ballot.

        Args:
            rank_ballot_data (dict[str, list]): Dictionary storing ballot data.
            idx (int): Index of ballot.
            rank_ballot (RankBallot): Ballot.
            candidates_cast (list[str]): List of candidates who have received votes.
            num_ballots (int): Total number of ballots.
        """

        if rank_ballot.ranking is None:
            return

        for j, cand_set in enumerate(rank_ballot.ranking):
            for c in cand_set:
                if self.candidates != tuple():
                    if c not in self.candidates:
                        raise ProfileError(
                            f"Candidate {c} found in ballot {rank_ballot} but not in "
                            f"candidate list {self.candidates}."
                        )
                if rank_ballot.weight > 0 and c not in candidates_cast:
                    candidates_cast.append(c)
            if f"Ranking_{j+1}" not in rank_ballot_data:
                assert self.max_ranking_length is not None

                if self.max_ranking_length > 0:
                    raise ProfileError(
                        f"Max ranking length {self.max_ranking_length} given but "
                        f"ballot {rank_ballot} has length at least {j+1}."
                    )
                rank_ballot_data[f"Ranking_{j+1}"] = [frozenset("~")] * num_ballots

            rank_ballot_data[f"Ranking_{j+1}"][idx] = cand_set

    def __update_rank_ballot_data_attrs(
        self,
        rank_ballot_data: dict[str, list],
        idx: int,
        rank_ballot: RankBallot,
        candidates_cast: list[str],
        num_ballots: int,
    ):
        """
        Update ballot data from a rank ballot.

        Args:
            rank_ballot_data (dict[str, list]): Dictionary storing ballot data.
            idx (int): Index of ballot.
            rank_ballot (RankBallot): Ballot.
            candidates_cast (list[str]): List of candidates who have received votes.
            num_ballots (int): Total number of ballots.
        """
        rank_ballot_data["Weight"][idx] = rank_ballot.weight

        if rank_ballot.voter_set != frozenset():
            rank_ballot_data["Voter Set"][idx] = rank_ballot.voter_set

        if rank_ballot.ranking is not None:
            self.__update_ballot_ranking_data(
                rank_ballot_data=rank_ballot_data,
                idx=idx,
                rank_ballot=rank_ballot,
                candidates_cast=candidates_cast,
                num_ballots=num_ballots,
            )

    def __init_rank_ballot_data(
        self, rank_ballots: Sequence[RankBallot]
    ) -> Tuple[int, dict[str, list]]:
        """
        Create the ballot data objects.

        Args:
            rank_ballots (Sequence[RankBallot,...]): Tuple of ballots.

        Returns:
            Tuple[int, dict[str, list]]: num_ballots, rank_ballot_data

        """
        num_ballots = len(rank_ballots)

        rank_ballot_data: dict[str, list] = {
            "Weight": [np.nan] * num_ballots,
            "Voter Set": [set()] * num_ballots,
        }
        assert self.max_ranking_length is not None

        if self.max_ranking_length > 0:
            rank_ballot_data.update(
                {
                    f"Ranking_{i+1}": [frozenset("~")] * num_ballots
                    for i in range(self.max_ranking_length)
                }
            )
        return num_ballots, rank_ballot_data

    def __init_formatted_rank_df(
        self,
        rank_ballot_data: dict[str, list],
    ) -> pd.DataFrame:
        """
        Create a pandas dataframe from the ballot data.

        Args:
            rank_ballot_data (dict[str, list]): Dictionary storing ballot data.

        Returns:
            pd.DataFrame: Dataframe of profile.
        """
        df = pd.DataFrame(rank_ballot_data)
        col_order = [c for c in df.columns if "Ranking_" in c] + [
            "Voter Set",
            "Weight",
        ]

        df = df[col_order]
        df.index.name = "Ballot Index"
        return df

    def _init_from_rank_ballots(
        self, ballots: Sequence[RankBallot]
    ) -> tuple[pd.DataFrame, tuple[str, ...]]:
        """
        Create the pandas dataframe representation of the profile.

        Args:
            ballots (Sequence[RankBallot,...]): Sequence of ballots.

        Returns:
            tuple[pd.DataFrame, tuple[str, ...]]: df, candidates_cast

        """
        # `rank_ballot_data` sends {Weight, Voter Set} keys to a list to be
        # indexed in the same order as the output df containing information
        # for each ballot. So rank_ballot_data[<weight>][<index>] is the weight value for
        # the ballot at index <index> in the df.
        num_ballots, rank_ballot_data = self.__init_rank_ballot_data(ballots)

        candidates_cast: list[str] = []

        for i, b in enumerate(ballots):
            self.__update_rank_ballot_data_attrs(
                rank_ballot_data=rank_ballot_data,
                idx=i,
                rank_ballot=b,
                candidates_cast=candidates_cast,
                num_ballots=num_ballots,
            )

        df = self.__init_formatted_rank_df(
            rank_ballot_data=rank_ballot_data,
        )

        return (
            df,
            tuple(candidates_cast),
        )

    def __validate_init_rank_df_params(self, df: pd.DataFrame) -> None:
        """
        Validate that the correct params were passed to the init method when constructing
        from a dataframe.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Raises:
            ProfileError: max_ranking_length must be provided.
            ProfileError: Candidates must be provided.
        """
        boiler_plate = (
            "When providing a dataframe and no ballot list to the init method, "
        )
        if len(df) == 0:
            return

        if self.max_ranking_length == 0:
            raise ProfileError(
                boiler_plate + "max_ranking_length must be provided and be non-zero."
            )

    def __validate_init_rank_df(self, df: pd.DataFrame) -> None:
        """
        Validate that the df passed to the init method is of valid type.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Raises:
            ProfileError: Ranking column is missing.
            ProfileError: Weight column is missing.
            ProfileError: Voter set column is missing.
            ProfileError: Index column is misformatted.

        """
        if "Weight" not in df.columns:
            raise ProfileError(f"Weight column not in dataframe: {df.columns}")
        if "Voter Set" not in df.columns:
            raise ProfileError(f"Voter Set column not in dataframe: {df.columns}")
        if df.index.name != "Ballot Index":
            raise ProfileError(f"Index not named 'Ballot Index': {df.index.name}")
        assert self.max_ranking_length is not None
        if any(
            f"Ranking_{i+1}" not in df.columns for i in range(self.max_ranking_length)
        ):
            for i in range(self.max_ranking_length):
                if f"Ranking_{i+1}" not in df.columns:
                    raise ProfileError(
                        f"Ranking column 'Ranking_{i+1}' not in dataframe: {df.columns}"
                    )

    def __find_candidates_cast_from_init_rank_df(
        self, df: pd.DataFrame
    ) -> tuple[str, ...]:
        """
        Compute which candidates received votes from the df and set the candidates_cast and
        candidates attr.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Returns:
            tuple[str]: Candidates cast.
        """

        mask = df["Weight"] > 0

        candidates_cast: set[str] = set()

        ranking_cols = [c for c in df.columns if c.startswith("Ranking_")]
        sets = df.loc[mask, ranking_cols].to_numpy().ravel()
        candidates_cast |= set().union(*sets)

        candidates_cast.discard("~")
        return tuple(candidates_cast)

    def _init_from_rank_df(
        self, df: pd.DataFrame
    ) -> tuple[pd.DataFrame, tuple[str, ...]]:
        """
        Validate the dataframe and determine the candidates cast.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Returns
            tuple[pd.DataFrame, tuple[str]]: df, candidates_cast
        """
        self.__validate_init_rank_df_params(df)
        self.__validate_init_rank_df(df)
        candidates_cast = self.__find_candidates_cast_from_init_rank_df(df)

        if len(df) == 0:
            self.max_ranking_length = 0

        return df, candidates_cast

    def _find_max_ranking_length(self) -> int:
        """
        Compute and set the maximum ranking length of the profile.

        Returns:
            int: Max ranking length.

        """
        if self.max_ranking_length == 0 or self.max_ranking_length is None:
            return len([c for c in self.df.columns if "Ranking_" in c])

        return self.max_ranking_length

    @cached_property
    def ballots(self: RankProfile) -> tuple[RankBallot, ...]:
        """
        Compute the ballot tuple as a cached property.
        """
        assert self.max_ranking_length is not None

        # TODO do this with map?
        computed_ballots = [RankBallot()] * len(self.df)
        for i, (_, b_row) in enumerate(self.df.iterrows()):
            computed_ballots[i] = convert_row_to_rank_ballot(
                b_row,
                self.max_ranking_length,
            )
        return tuple(computed_ballots)

    def __add__(self, other) -> RankProfile:
        """
        Add two PreferenceProfiles by combining their ballot lists.
        """
        if not isinstance(other, RankProfile):
            raise TypeError(
                "Unsupported operand type. Must be an instance of RankProfile."
            )

        assert (
            self.max_ranking_length is not None and other.max_ranking_length is not None
        )
        max_ranking_length = max([self.max_ranking_length, other.max_ranking_length])
        candidates = list(set(self.candidates).union(other.candidates))

        df_1 = self.df.copy()
        df_2 = other.df.copy()

        if self.max_ranking_length < max_ranking_length:
            for i in range(self.max_ranking_length, max_ranking_length):
                df_1[f"Ranking_{i + 1}"] = [frozenset("~")] * len(df_1)
        if other.max_ranking_length < max_ranking_length:
            for i in range(other.max_ranking_length, max_ranking_length):
                df_2[f"Ranking_{i + 1}"] = [frozenset("~")] * len(df_2)

        new_df = pd.concat([df_1, df_2], ignore_index=True)
        new_df.index.name = "Ballot Index"
        ranking_cols = [c for c in new_df.columns if "Ranking_" in c]
        new_df[ranking_cols] = new_df[ranking_cols].astype("object")
        new_df = new_df[
            [f"Ranking_{i + 1}" for i in range(max_ranking_length)]
            + ["Weight", "Voter Set"]
        ]

        return RankProfile(
            candidates=candidates,
            df=new_df,
            max_ranking_length=max_ranking_length,
        )

    def group_ballots(self) -> RankProfile:
        """
        Groups ballots by rankings and updates weights. Retains voter sets, but
        loses ballot indices.

        Returns:
            RankProfile: A RankProfile object with grouped ballot list.
        """
        empty_df = pd.DataFrame(columns=["Voter Set", "Weight"], dtype=np.float64)
        empty_df.index.name = "Ballot Index"

        if len(self.df) == 0:
            return RankProfile(
                candidates=self.candidates,
                max_ranking_length=self.max_ranking_length,
            )

        ranking_cols = [c for c in self.df.columns if "Ranking_" in c]
        group_df = self.df.groupby(ranking_cols, dropna=False)
        new_df = group_df.aggregate(
            {
                "Weight": "sum",
                "Voter Set": (lambda sets: set().union(*sets)),
            }
        ).reset_index()

        new_df.index.name = "Ballot Index"

        return RankProfile(
            df=new_df,
            candidates=self.candidates,
            max_ranking_length=self.max_ranking_length,
        )

    def __eq__(self, other):
        if not isinstance(other, RankProfile):
            return False

        if self.max_ranking_length != other.max_ranking_length:
            return False

        return super().__eq__(other)

    def __str__(self) -> str:

        repr_str = "RankProfile\n"
        repr_str += f"Maximum ranking length: {self.max_ranking_length}\n"

        repr_str += (
            f"Candidates: {self.candidates}\n"
            f"Candidates who received votes: {self.candidates_cast}\n"
            f"Total number of Ballot objects: {self.num_ballots}\n"
            f"Total weight of Ballot objects: {self.total_ballot_wt}\n"
        )

        return repr_str

    __repr__ = __str__

    def __to_rank_csv_header(
        self, candidate_mapping: dict[str, str], include_voter_set: bool
    ) -> list[list]:
        """
        Construct the header rows for the PrefProfile a custom CSV format.

        Args:
            include_voter_set (bool): Whether or not to include the voter set of each
                ballot.
        """
        header = [
            ["VoteKit RankProfile"],
            ["Candidates"],
            [f"({c}:{cand_label})" for c, cand_label in candidate_mapping.items()],
        ]
        header += [["Max Ranking Length"], [str(self.max_ranking_length)]]
        header += [["Includes Voter Set"], [str(include_voter_set)]]
        header += [["="] * 10]

        return header

    def __to_rank_csv_ranking_list(
        self, rank_ballot: RankBallot, candidate_mapping: dict[str, str]
    ) -> list:
        """
        Create the list of ranking data for a ballot in the profile.

        Args:
            rank_ballot (RankBallot): Ballot.
            candidate_mapping (dict[str, int]): Mapping candidate names to integers.

        """
        assert self.max_ranking_length is not None

        if rank_ballot.ranking is not None:
            ranking_list = [
                (
                    set([candidate_mapping[c] for c in cand_set])
                    if cand_set != frozenset()
                    else "{}"
                )
                for cand_set in rank_ballot.ranking
            ]
            if len(ranking_list) != self.max_ranking_length:
                ranking_list += [""] * (self.max_ranking_length - len(ranking_list))

            return ranking_list

        return [""] * self.max_ranking_length

    def __to_rank_csv_ballot_row(
        self,
        ballot: RankBallot,
        include_voter_set: bool,
        candidate_mapping: dict[str, str],
        weight_precision: int,
    ) -> list[list]:
        """
        Create the row for a ballot in the profile.

        Args:
            ballot (Ballot): Ballot.
            include_voter_set (bool): Whether or not to include the voter set of each
                ballot.
            candidate_mapping (dict[str, int]): Mapping candidate names to integers.
            weight_precision (int): Number of decimals to round float weights to.

        """
        row = self.__to_rank_csv_ranking_list(ballot, candidate_mapping)
        row += ["&"]

        row += [round(ballot.weight, weight_precision), "&"]

        if include_voter_set:
            row += [v for v in sorted(ballot.voter_set)]

        return row

    def __to_rank_csv_data_column_names(
        self, include_voter_set: bool, candidate_mapping: dict[str, str]
    ) -> list:
        """
        Create the data column header.

        Args:
            include_voter_set (bool): Whether or not to include the voter set of each
                ballot.
            candidate_mapping (dict[str, str]): Maps candidate names to prefixes.
        """
        assert self.max_ranking_length is not None
        data_col_names = [f"Ranking_{i+1}" for i in range(self.max_ranking_length)]
        data_col_names += ["&", "Weight", "&"]

        if include_voter_set:
            data_col_names += ["Voter Set"]

        return data_col_names

    def to_csv(
        self,
        fpath: Union[str, PathLike, Path],
        include_voter_set: bool = False,
        weight_precision: int = 2,
    ):
        """
        Saves PreferenceProfile to a custom CSV format.

        Args:
            fpath (Union[str, PathLike, Path]): Path to the saved csv.
            include_voter_set (bool, optional): Whether or not to include the voter set of each
                ballot. Defaults to False.
            weight_precision (int): Number of decimals to round float weights to. Defaults to 2.
        Raises:
            ProfileError: Cannot write a profile with no ballots to a csv.
            ValueError: File path must be provided.
        """
        if fpath == "":
            raise ValueError("File path must be provided.")

        if len(self.ballots) == 0:
            raise ProfileError("Cannot write a profile with no ballots to a csv.")

        prefix_idx = 1
        candidate_mapping = {c: c[:prefix_idx] for c in self.candidates}
        while len(set(candidate_mapping.values())) < len(candidate_mapping.values()):
            prefix_idx += 1
            candidate_mapping = {c: c[:prefix_idx] for c in self.candidates}

        header = self.__to_rank_csv_header(candidate_mapping, include_voter_set)
        data_col_names = self.__to_rank_csv_data_column_names(
            include_voter_set, candidate_mapping
        )
        ballot_rows = [
            self.__to_rank_csv_ballot_row(
                b, include_voter_set, candidate_mapping, weight_precision
            )
            for b in self.ballots
        ]
        rows = header + [data_col_names] + ballot_rows

        with open(
            str(fpath),
            "w",
            newline="",
            encoding="utf-8",
        ) as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(rows)

    @classmethod
    def from_csv(cls, fpath: Union[str, PathLike, Path]) -> PreferenceProfile:
        """
        Creates a PreferenceProfile from a csv, formatted from the ``to_csv`` method.

        Args:
            fpath (Union[str, PathLike, Path]): Path to csv.

        Raises:
            ValueError: If csv is improperly formatted for VoteKit.
            ProfileError: If read profile has no rankings or scores.

        """
        with open(str(fpath), "r") as file:
            reader = csv.reader(file)
            csv_data = list(reader)

        _validate_rank_csv_format(csv_data)

        (
            inv_candidate_mapping,
            max_ranking_length,
            includes_voter_set,
            break_indices,
        ) = _parse_profile_data_from_rank_csv(csv_data)

        ballots = [
            _parse_ballot_from_rank_csv(
                row,
                includes_voter_set,
                break_indices,
                inv_candidate_mapping,
            )
            for row in csv_data[9:]
        ]

        return cls(
            ballots=tuple(ballots),
            candidates=tuple(inv_candidate_mapping.values()),
            max_ranking_length=max_ranking_length,
        )


class ScoreProfile(PreferenceProfile):

    def __init__(
        self,
        *,
        ballots: Sequence[ScoreBallot] = tuple(),
        candidates: Sequence[str] = tuple(),
        max_ranking_length: Optional[int] = None,
        df: pd.DataFrame = pd.DataFrame(),
    ):
        self.candidates = tuple(candidates)

        if df.equals(pd.DataFrame()):
            (
                self.df,
                self.candidates_cast,
            ) = self._init_from_score_ballots(ballots)

            if self.candidates == tuple():
                self.candidates = self.candidates_cast

        else:
            self.df, self.candidates_cast = self._init_from_score_df(df)

        super().__init__(
            candidates=self.candidates,
            candidates_cast=self.candidates_cast,
            df=self.df,
        )

    def __update_ballot_scores_data(
        self,
        score_ballot_data: dict[str, list],
        idx: int,
        ballot: ScoreBallot,
        candidates_cast: list[str],
        num_ballots: int,
    ) -> None:
        """
        Update the score data from a ballot.

        Args:
            ballot_data (dict[str, list]): Dictionary storing ballot data.
            idx (int): Index of ballot.
            ballot (ScoreBallot): Ballot.
            candidates_cast (list[str]): List of candidates who have received votes.
            num_ballots (int): Total number of ballots.
        """
        if ballot.scores is None:
            return

        for c, score in ballot.scores.items():
            if ballot.weight > 0 and c not in candidates_cast:
                candidates_cast.append(c)

            if c not in score_ballot_data:
                if self.candidates:
                    raise ProfileError(
                        f"Candidate {c} found in ballot {ballot} but not in "
                        f"candidate list {self.candidates}."
                    )
                score_ballot_data[c] = [np.nan] * num_ballots
            score_ballot_data[c][idx] = score

    def __update_score_ballot_data_attrs(
        self,
        score_ballot_data: dict[str, list],
        idx: int,
        ballot: ScoreBallot,
        candidates_cast: list[str],
        num_ballots: int,
    ) -> None:
        """
        Update all ballot data from a ballot.

        Args:
            ballot_data (dict[str, list]): Dictionary storing ballot data.
            idx (int): Index of ballot.
            ballot (ScoreBallot): Ballot.
            candidates_cast (list[str]): List of candidates who have received votes.
            num_ballots (int): Total number of ballots.
        """
        score_ballot_data["Weight"][idx] = ballot.weight

        if ballot.voter_set != frozenset():
            score_ballot_data["Voter Set"][idx] = ballot.voter_set

        if ballot.scores is not None:
            self.__update_ballot_scores_data(
                score_ballot_data=score_ballot_data,
                idx=idx,
                ballot=ballot,
                candidates_cast=candidates_cast,
                num_ballots=num_ballots,
            )

    def __init_score_ballot_data(
        self, ballots: Sequence[ScoreBallot]
    ) -> Tuple[int, dict[str, list]]:
        """
        Create the ballot data objects.

        Args:
            ballots (Sequence[ScoreBallot,...]): Tuple of ballots.

        Returns:
            Tuple[int, dict[str, list]]: num_ballots, score_ballot_data

        """
        num_ballots = len(ballots)

        score_ballot_data: dict[str, list] = {
            "Weight": [np.nan] * num_ballots,
            "Voter Set": [set()] * num_ballots,
        }

        if self.candidates != tuple():
            score_ballot_data.update(
                {c: [np.nan] * num_ballots for c in self.candidates}
            )

        return num_ballots, score_ballot_data

    def __init_formatted_score_df(
        self,
        score_ballot_data: dict[str, list],
        candidates_cast: list[str],
    ) -> pd.DataFrame:
        """
        Create a pandas dataframe from the ballot data.

        Args:
            score_ballot_data (dict[str, list]): Dictionary storing ballot data.
            candidates_cast (list[str]): List of candidates who received votes.

        Returns:
            pd.DataFrame: Dataframe of profile.
        """
        df = pd.DataFrame(score_ballot_data)
        temp_col_order = [
            "Voter Set",
            "Weight",
        ]

        col_order = list(self.candidates) + temp_col_order

        if self.candidates == tuple():
            remaining_cands = set(candidates_cast) - set(df.columns)
            empty_df_cols = np.full((len(df), len(remaining_cands)), np.nan)
            df[list(remaining_cands)] = empty_df_cols

            col_order = (
                sorted([c for c in df.columns if c not in temp_col_order])
                + temp_col_order
            )

        df = df[col_order]
        df.index.name = "Ballot Index"
        return df

    def _init_from_score_ballots(
        self, ballots: Sequence[ScoreBallot]
    ) -> tuple[pd.DataFrame, tuple[str, ...]]:
        """
        Create the pandas dataframe representation of the profile.

        Args:
            ballots (Sequence[ScoreBallot,...]): Tuple of ballots.

        Returns:
            tuple[pd.DataFrame, tuple[str, ...]]: df, candidates_cast

        """
        # `score_ballot_data` sends {Weight, Voter Set} keys to a list to be
        # indexed in the same order as the output df containing information
        # for each ballot. So ballot_data[<weight>][<index>] is the weight value for
        # the ballot at index <index> in the df.
        num_ballots, score_ballot_data = self.__init_score_ballot_data(ballots)

        candidates_cast: list[str] = []

        for i, b in enumerate(ballots):

            self.__update_score_ballot_data_attrs(
                score_ballot_data=score_ballot_data,
                idx=i,
                ballot=b,
                candidates_cast=candidates_cast,
                num_ballots=num_ballots,
            )

        df = self.__init_formatted_score_df(
            score_ballot_data=score_ballot_data,
            candidates_cast=candidates_cast,
        )
        return (
            df,
            tuple(candidates_cast),
        )

    def __validate_init_score_df_params(self, df: pd.DataFrame) -> None:
        """
        Validate that the correct params were passed to the init method when constructing
        from a dataframe.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Raises:
            ValueError: One of contains_rankings and contains_scores must be True.
            ValueError: If contains_rankings is True, max_ranking_length must be provided.
            ValueError: Candidates must be provided.
        """
        boiler_plate = (
            "When providing a dataframe and no ballot list to the init method, "
        )
        if len(df) == 0:
            return

        if self.candidates == tuple():
            raise ProfileError(boiler_plate + "candidates must be provided.")

    def __validate_init_score_df(self, df: pd.DataFrame) -> None:
        """
        Validate that the df passed to the init method is of valid type.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Raises:
            ValueError: Candidate column is missing.
            ValueError: Ranking column is missing.
            ValueError: Weight column is missing.
            ValueError: Voter set column is missing.
            ValueError: Index column is misformatted.

        """
        if "Weight" not in df.columns:
            raise ProfileError(f"Weight column not in dataframe: {df.columns}")
        if "Voter Set" not in df.columns:
            raise ProfileError(f"Voter Set column not in dataframe: {df.columns}")
        if df.index.name != "Ballot Index":
            raise ProfileError(f"Index not named 'Ballot Index': {df.index.name}")
        if any(c not in df.columns for c in self.candidates):
            for c in self.candidates:
                if c not in df.columns:
                    raise ProfileError(
                        f"Candidate column '{c}' not in dataframe: {df.columns}"
                    )

    def __find_candidates_cast_from_init_score_df(
        self, df: pd.DataFrame
    ) -> tuple[str, ...]:
        """
        Compute which candidates received votes from the df and set the candidates_cast and
        candidates attr.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Returns:
            tuple[str]: Candidates cast.
        """

        mask = df["Weight"] > 0

        candidates_cast: set[str] = set()

        positive = df.loc[mask, list(self.candidates)].gt(0).any()
        # .any() applies along the columns, so we get a boolean series where the
        # value is True the candidate has any positive score the column
        candidates_cast |= set(positive[positive].index)

        return tuple(candidates_cast)

    def _init_from_score_df(
        self, df: pd.DataFrame
    ) -> tuple[pd.DataFrame, tuple[str, ...]]:
        """
        Validate the dataframe and determine the candidates cast.

        Args:
            df (pd.DataFrame): Dataframe representation of ballots.

        Returns
            tuple[pd.DataFrame, tuple[str]]: df, candidates_cast
        """
        self.__validate_init_score_df_params(df)
        self.__validate_init_score_df(df)
        candidates_cast = self.__find_candidates_cast_from_init_score_df(df)

        return df, candidates_cast

    @cached_property
    def ballots(self: ScoreProfile) -> tuple[ScoreBallot, ...]:
        """
        Compute the ballot tuple as a cached property.
        """
        # TODO with map?
        computed_ballots = [ScoreBallot()] * len(self.df)
        for i, (_, b_row) in enumerate(self.df.iterrows()):
            computed_ballots[i] = convert_row_to_score_ballot(
                b_row, tuple(self.candidates)
            )
        return tuple(computed_ballots)

    def __add__(self, other):
        """
        Add two PreferenceProfiles by combining their ballot lists.
        """
        if not isinstance(other, ScoreProfile):
            raise TypeError(
                "Unsupported operand type. Must be an instance of ScoreProfile."
            )

        df_1 = self.df.copy()
        df_2 = other.df.copy()

        cand1 = set(self.candidates)
        cand2 = set(other.candidates)
        for cand in cand2 - cand1:
            df_1[cand] = [np.nan] * len(df_1)
        for cand in cand1 - cand2:
            df_2[cand] = [np.nan] * len(df_2)

        new_df = pd.concat([df_1, df_2], ignore_index=True)
        new_df.index.name = "Ballot Index"

        new_candidates = sorted(set(self.candidates).union(other.candidates))
        new_df = new_df[new_candidates + ["Weight", "Voter Set"]]

        return ScoreProfile(
            candidates=new_candidates,
            df=new_df,
        )

    def group_ballots(self) -> ScoreProfile:
        """
        Groups ballots by scores and updates weights. Retains voter sets, but
        loses ballot ids.

        Returns:
            ScoreProfile: A ScoreProfile object with grouped ballot list.
        """
        empty_df = pd.DataFrame(columns=["Voter Set", "Weight"], dtype=np.float64)
        empty_df.index.name = "Ballot Index"

        if len(self.df) == 0:
            return ScoreProfile(
                candidates=self.candidates,
            )

        non_group_cols = ["Weight", "Voter Set"]
        cand_cols = [c for c in self.df.columns if c not in non_group_cols]

        group_df = self.df.groupby(cand_cols, dropna=False)
        new_df = group_df.aggregate(
            {
                "Weight": "sum",
                "Voter Set": (lambda sets: set().union(*sets)),
            }
        ).reset_index()

        new_df.index.name = "Ballot Index"

        return ScoreProfile(
            df=new_df,
            candidates=self.candidates,
        )

    def __eq__(self, other):
        if not isinstance(other, ScoreProfile):
            return False

        return super().__eq__(other)

    def __str__(self) -> str:

        repr_str = "ScoreProfile\n"

        repr_str += (
            f"Candidates: {self.candidates}\n"
            f"Candidates who received votes: {self.candidates_cast}\n"
            f"Total number of Ballot objects: {self.num_ballots}\n"
            f"Total weight of Ballot objects: {self.total_ballot_wt}\n"
        )

        return repr_str

    __repr__ = __str__

    def __to_score_csv_header(
        self, candidate_mapping: dict[str, str], include_voter_set: bool
    ) -> list[list]:
        """
        Construct the header rows for the PrefProfile a custom CSV format.

        Args:
            include_voter_set (bool): Whether or not to include the voter set of each
                ballot.
        """
        header = [
            ["VoteKit ScoreProfile"],
            ["Candidates"],
            [f"({c}:{cand_label})" for c, cand_label in candidate_mapping.items()],
        ]
        header += [["Includes Voter Set"], [str(include_voter_set)]]
        header += [["="] * 10]

        return header

    def __to_score_csv_score_list(self, score_ballot: ScoreBallot) -> list:
        """
        Create the list of score data for a ballot in the profile.

        Args:
            ballot (Ballot): Ballot.

        """
        if score_ballot.scores is not None:
            return [float(score_ballot.scores.get(c, 0)) for c in self.candidates]

        return [""] * len(self.candidates)

    def __to_score_csv_ballot_row(
        self,
        score_ballot: ScoreBallot,
        include_voter_set: bool,
        weight_precision: int,
    ) -> list[list]:
        """
        Create the row for a ballot in the profile.

        Args:
            score_ballot (ScoreBallot): Ballot.
            include_voter_set (bool): Whether or not to include the voter set of each
                ballot.
            weight_precision (int): Number of decimals to round float weights to.


        """
        row = self.__to_score_csv_score_list(score_ballot)
        row += ["&"]

        row += [round(score_ballot.weight, weight_precision), "&"]

        if include_voter_set:
            row += [v for v in sorted(score_ballot.voter_set)]

        return row

    def __to_score_csv_data_column_names(
        self, include_voter_set: bool, candidate_mapping: dict[str, str]
    ) -> list:
        """
        Create the data column header.

        Args:
            include_voter_set (bool): Whether or not to include the voter set of each
                ballot.
            candidate_mapping (dict[str, str]): Maps candidate names to prefixes.
        """
        data_col_names = [f"{cand_label}" for cand_label in candidate_mapping.values()]
        data_col_names += ["&", "Weight", "&"]

        if include_voter_set:
            data_col_names += ["Voter Set"]

        return data_col_names

    def to_csv(
        self,
        fpath: Union[str, PathLike, Path],
        include_voter_set: bool = False,
        weight_precision: int = 2,
    ):
        """
        Saves PreferenceProfile to a custom CSV format.

        Args:
            fpath (Union[str, PathLike, Path]): Path to the saved csv.
            include_voter_set (bool, optional): Whether or not to include the voter set of each
                ballot. Defaults to False.
            weight_precision (int): Number of decimals to round float weights to. Defaults to 2.

        Raises:
            ProfileError: Cannot write a profile with no ballots to a csv.
            ValueError: File path must be provided.
        """
        if fpath == "":
            raise ValueError("File path must be provided.")

        if len(self.ballots) == 0:
            raise ProfileError("Cannot write a profile with no ballots to a csv.")

        prefix_idx = 1
        candidate_mapping = {c: c[:prefix_idx] for c in self.candidates}
        while len(set(candidate_mapping.values())) < len(candidate_mapping.values()):
            prefix_idx += 1
            candidate_mapping = {c: c[:prefix_idx] for c in self.candidates}

        header = self.__to_score_csv_header(candidate_mapping, include_voter_set)
        data_col_names = self.__to_score_csv_data_column_names(
            include_voter_set, candidate_mapping
        )
        ballot_rows = [
            self.__to_score_csv_ballot_row(b, include_voter_set, weight_precision)
            for b in self.ballots
        ]
        rows = header + [data_col_names] + ballot_rows

        with open(
            str(fpath),
            "w",
            newline="",
            encoding="utf-8",
        ) as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(rows)

    @classmethod
    def from_csv(cls, fpath: Union[str, PathLike, Path]) -> PreferenceProfile:
        """
        Creates a PreferenceProfile from a csv, formatted from the ``to_csv`` method.

        Args:
            fpath (str): Path to csv.

        Raises:
            ValueError: If csv is improperly formatted for VoteKit.
            ProfileError: If read profile has no rankings or scores.

        """
        with open(str(fpath), "r") as file:
            reader = csv.reader(file)
            csv_data = list(reader)

        _validate_score_csv_format(csv_data)

        (
            inv_candidate_mapping,
            includes_voter_set,
            break_indices,
        ) = _parse_profile_data_from_score_csv(csv_data)

        ballots = [
            _parse_ballot_from_score_csv(
                row,
                includes_voter_set,
                break_indices,
                inv_candidate_mapping,
            )
            for row in csv_data[7:]
        ]

        return cls(
            ballots=tuple(ballots),
            candidates=tuple(inv_candidate_mapping.values()),
        )



from votekit.pref_profile import RankProfile, rank_profile_to_ranking_dict
import numpy as np
from typing import Union, Optional, Sequence
from scipy.optimize import linprog
from scipy.sparse import identity, kron, vstack, csr_matrix
from scipy.stats import kendalltau


def emd_via_scipy_linear_program(
    source_distribution: np.ndarray,
    target_distribution: np.ndarray,
    cost_matrix: np.ndarray,
) -> float:
    """
    Compute the Earth Mover's Distance (EMD) between two discrete distributions using
    a linear programming formulation of the optimal transport problem.

    Classical Formulation:
        minimize ⟨cost_matrix, transport_plan⟩
        subject to:
            transport_plan @ 1 = source_distribution
            transport_plan.T @ 1 = target_distribution
            transport_plan >= 0

    Linear Program:
        Suppose that 'a' is the source distribution, 'b' is the target distribution, 'C' is the
        cost matrix, and 'X' is the transport plan. Let 'x' be the row-major vectorization of 'X'
        and let 'c' be the vectorization of 'C'.

        Then let 'A' be the constraint matrix formed by the Kronecker products of identity
        matrices and the ones vector, so

        A = [I_n ⊗  1_m^T, 1_n^T ⊗ I_m]

        where 'I_n' is the identity matrix of size 'n' (number of sources) and 'I_m' is the identity
        matrix of size 'm' (number of targets). The linear program can be expressed as:

        minimize c^T x

        subject to:
            Ax = y

        where 'y' is the concatenation of the source and target distributions 'y = [a, b]^T'.

    Args:
        source_distribution (np.ndarray): Mass distribution of the source (length n).
        target_distribution (np.ndarray): Mass distribution of the target (length m).
        cost_matrix (np.ndarray): n×m matrix of transportation costs.

    Returns:
        float: The computed Earth Mover's Distance (minimum transport cost).
    """
    nonzero_source_mask = source_distribution > 0
    nonzero_target_mask = target_distribution > 0

    if not nonzero_source_mask.any() and not nonzero_target_mask.any():
        return 0.0

    if (source_distribution < 0).any() or (target_distribution < 0).any():
        raise ValueError(
            "Negative entries in source or target distributions are not allowed."
        )

    # Trim to only nonzero entries to reduce LP size
    trimmed_source = source_distribution[nonzero_source_mask]
    trimmed_target = target_distribution[nonzero_target_mask]

    # NOTE: np.ix_ is used to create a meshgrid for indexing so you grab the correct submatrix
    trimmed_cost_matrix = cost_matrix[np.ix_(nonzero_source_mask, nonzero_target_mask)]
    num_sources, num_targets = trimmed_cost_matrix.shape

    # For each source, we need to ensure that the total mass is fully received by the targets
    # Suppose n=2 sources, m=3 targets, and transport plan:
    #
    #     X = [[x11, x12, x13],
    #          [x21, x22, x23]]
    #
    # Flattened row-major:
    #     x = [x11, x12, x13, x21, x22, x23]   # length n*m = 6
    #
    # Now with the Kronecker product, we get the constraints:
    # Row constraints: I_n ⊗ [1 1 1]   → sums each source's row:
    #     [[1 1 1 0 0 0],               # x11 + x12 + x13 = a1
    #      [0 0 0 1 1 1]]               # x21 + x22 + x23 = a2
    #
    # Col constraints: [1 1] ⊗ I_m     → sums each target's column:
    #     [[1 0 0 1 0 0],               # x11 + x21 = b1
    #      [0 1 0 0 1 0],               # x12 + x22 = b2
    #      [0 0 1 0 0 1]]               # x13 + x23 = b3
    row_constraints = kron(
        identity(num_sources, format="csr"),
        csr_matrix(np.ones((1, num_targets))),
    )
    col_constraints = kron(
        csr_matrix(np.ones((1, num_sources))),
        identity(num_targets, format="csr"),
    )

    # Matrix formulation of the constraints:
    A_eq = vstack([row_constraints, col_constraints], format="csr")

    # NOTE: np.r_ is used to concatenate the trimmed source and target distributions
    b_eq = np.r_[trimmed_source, trimmed_target]

    # Solve LP
    result = linprog(
        trimmed_cost_matrix.ravel(),
        A_eq=A_eq,
        b_eq=b_eq,
        bounds=(0, None),
        method="highs",
    )

    if not result.success:
        raise RuntimeError(f"linprog failed: {result.message}")

    ret = result.fun
    if ret is None:
        raise RuntimeError("linprog returned None as the optimal value.")

    assert ret is not None
    return ret


def __vaildate_ranking_distance_inputs(
    ranking1: Sequence[int], ranking2: Sequence[int], n_candidates: int
) -> tuple[set[int], set[int], set[int]]:
    """
    Validates the inputs for computing the distance between two rankings.

    Args:
        ranking1 (Sequence[int]): The first ranking as a sequence of candidate indices.
        ranking2 (Sequence[int]): The second ranking as a sequence of candidate indices.
        n_candidates (int): The total number of candidates in the rankings.

    Returns:
        tuple[set[int], set[int], set[int]]: A tuple containing three sets:
            - The set of candidates in the first ranking.
            - The set of candidates in the second ranking.
            - The union of both sets, representing all candidates in both rankings.

    Raises:
        ValueError: If the rankings contain duplicates or if the number of candidates jointly
            contained within the rankings exceeds the total number of candidates.
    """
    ranked1_set = set(ranking1)
    ranked2_set = set(ranking2)
    full_ranking_set = ranked1_set.union(ranked2_set)

    if len(ranked1_set) != len(ranking1):
        raise ValueError(
            f"The ranking {ranking1} contains duplicates and is not suitable for distance "
            "computation."
        )
    if len(ranked2_set) != len(ranking2):
        raise ValueError(
            f"The ranking {ranking2} contains duplicates and is not suitable for distance "
            "computation."
        )
    if len(full_ranking_set) > n_candidates:
        raise ValueError(
            "The number of candidates in the rankings exceeds the total number of candidates."
        )
    return ranked1_set, ranked2_set, full_ranking_set


def __compute_bubble_sort_distance(
    ranking1: Sequence[int], ranking2: Sequence[int], full_ranking_set: set[int]
) -> float:
    r"""
    Computes the bubble sort distance between two rankings.

    This function makes use of the Kendall tau correlation coefficient to compute the
    bubble sort distance, which is defined as the number of adjacent transpositions needed to
    transform one ranking into another. The formula for the Kendall tau statistic is given by:

    .. math::
        \tau = \frac{2(C - D}{n(n-1)}

    where :math:`C` is the number of concordant pairs, :math:`C` is the number of discordant pairs,
    and :math:`n` is the number of items being ranked. The bubble sort distance, :math:`D` is then
    computed as:

    .. math::
        \text{Bubble Sort Distance} = \frac{n(n-1)}{4} \cdot (1 - \tau)

    where :math:`n` is the number of candidates in the full ranking set.

    Args:
        ranking1 (Sequence[int]): The first ranking as a sequence of candidate indices.
        ranking2 (Sequence[int]): The second ranking as a sequence of candidate indices.
        full_ranking_set (set[int]): The set of all candidates in both rankings.

    Returns:
        float: The computed bubble sort distance between the two rankings.
    """
    if len(ranking1) == 0 or len(ranking2) == 0:
        return 0.0

    full_ranking_1 = list(ranking1) + [c for c in ranking2 if c not in set(ranking1)]
    full_ranking_2 = list(ranking2) + [c for c in ranking1 if c not in set(ranking2)]

    assert set(full_ranking_1) == set(full_ranking_2) == full_ranking_set

    ranking1_sort_idx = np.argsort(full_ranking_1)
    ranking2_sort_idx = np.argsort(full_ranking_2)

    tau, _ = kendalltau(ranking1_sort_idx, ranking2_sort_idx)
    full_ranking_count = len(full_ranking_set)
    t_value = (full_ranking_count * (full_ranking_count - 1)) / 2
    return int(round(t_value * (1 - float(tau)) / 2))  # type: ignore[return-value]


def compute_ranking_distance_on_ballot_graph(
    ranking1: Sequence[int],
    ranking2: Sequence[int],
    n_candidates: int,
):
    """
    Computes the distance between two rankings on a ballot graph.

    Args:
        ranking1 (Sequence[int]): The first ranking as a sequence of candidate indices.
        ranking2 (Sequence[int]): The second ranking as a sequence of candidate indices.
        n_candidates (int): The number of candidates in the rankings.

    Returns:
        float: The computed distance between the two rankings.
    """
    if n_candidates <= 0:
        raise ValueError("The number of candidates must be greater than zero.")
    if n_candidates == 1:
        return 0.0
    if ranking1 == ranking2:
        return 0.0

    ranked1_set, ranked2_set, full_ranking_set = __vaildate_ranking_distance_inputs(
        ranking1, ranking2, n_candidates
    )

    bubble_sort_distance = __compute_bubble_sort_distance(
        ranking1, ranking2, full_ranking_set
    )

    insertion_deletion_set = ranked1_set.symmetric_difference(ranked2_set)
    insertion_deletion_distance = len(insertion_deletion_set) / 2

    if (
        ranked1_set.intersection(ranked2_set) == set()
        and len(ranking1) > 0
        and len(ranking2) > 0
    ):
        # Case where there might an insertion and a deletion at the end. Both are free, so we
        # credit them back 0.5 each.
        insertion_credit = 1 if len(full_ranking_set) == n_candidates else 0.0
    else:
        # Case where there might be a single insertion or deletion at the end which we need to
        # credit back.
        insertion_credit = 0.5 if len(full_ranking_set) == n_candidates else 0.0

    # Case were where two ballots equivalent to full rankings swap the last two candidates
    if (
        len(full_ranking_set) == n_candidates
        and len(ranking1) != 0
        and len(ranking2) != 0
        and ranking1[:-2] == ranking2[:-2]
    ):
        insertion_deletion_distance = 0.0
        insertion_credit = 0.0

    return bubble_sort_distance + insertion_deletion_distance - insertion_credit


def __build_simultaneous_profile_distribution(
    pp1: RankProfile, pp2: RankProfile
) -> dict[tuple[int, ...], tuple[float, float]]:
    """
    Builds a simultaneous distribution of two preference profiles, where each key is a tuple
    representing a ranking and the value is a tuple containing the weights from each of the
    profiles in order. If the ranking is not present in one of the profiles, the weight is set to
    0.0.

    Args:
        pp1 (RankProfile): RankProfile for first profile.
        pp2 (RankProfile): RankProfile for second profile.

    Returns:
        dict[tuple[int, ...], tuple[float, float]]: A dictionary where keys are tuples of candidate
            indices representing rankings and values are tuples of weights from each profile
            in the order of pp1 and pp2.
    """
    profile1 = pp1.group_ballots()
    profile2 = pp2.group_ballots()

    cand_to_index_mapping = {
        cand: i for i, cand in enumerate(sorted(profile1.candidates))
    }

    profile_distribution_dict = dict()

    assert profile1.max_ranking_length is not None
    profile1_ranking_array = profile1.df[
        [f"Ranking_{i}" for i in range(1, profile1.max_ranking_length + 1)]
    ].to_numpy()
    profile1_wt_vector = profile1.df["Weight"].astype(float).to_numpy()
    profile1_wt_vector = profile1_wt_vector / np.sum(profile1_wt_vector)

    if (profile1_ranking_array == frozenset({})).any():
        raise ValueError(
            "The first profile contains an empty ranking, which is not allowed."
        )

    for idx, ranking_tuple in enumerate(profile1_ranking_array):
        tup = tuple(
            [
                cand_to_index_mapping[cand]
                for cand_set in ranking_tuple
                for cand in cand_set
                if cand != "~"
            ]
        )

        profile_distribution_dict[tup] = (float(profile1_wt_vector[idx]), 0.0)

    assert profile2.max_ranking_length is not None
    profile2_ranking_array = profile2.df[
        [f"Ranking_{i}" for i in range(1, profile2.max_ranking_length + 1)]
    ].to_numpy()
    profile2_wt_vector = profile2.df["Weight"].astype(float).to_numpy()
    profile2_wt_vector = profile2_wt_vector / np.sum(profile2_wt_vector)

    if (profile2_ranking_array == frozenset({})).any():
        raise ValueError(
            "The second profile contains an empty ranking, which is not allowed."
        )

    for idx, ranking_tuple in enumerate(profile2_ranking_array):
        tup = tuple(
            [
                cand_to_index_mapping[cand]
                for cand_set in ranking_tuple
                for cand in cand_set
                if cand != "~"
            ]
        )

        profile_distribution_dict[tup] = (
            profile_distribution_dict.get(tup, (0.0, 0.0))[0],
            float(profile2_wt_vector[idx]),
        )

    return profile_distribution_dict


def earth_mover_dist(pp1: RankProfile, pp2: RankProfile) -> float:
    """
    Computes the Earth Mover's Distance (EMD) between two preference profiles.
    Assumes both elections share the same candidates.

    Args:
        pp1 (RankProfile): RankProfile for first profile.
        pp2 (RankProfile): RankProfile for second profile.

    Returns:
        float: Earth Mover's Distance between two profiles.
    """
    if set(pp1.candidates) != set(pp2.candidates):
        raise ValueError("The two profiles must have the same candidates.")

    if not isinstance(pp1, RankProfile) or not isinstance(pp2, RankProfile):
        raise ValueError(
            "Both profiles must contain rankings to compute the Earth Mover's Distance."
        )
    if pp1.max_ranking_length != pp2.max_ranking_length:
        raise ValueError(
            "Both profiles must have the same maximum ranking length to compute the Earth Mover's "
            "Distance."
        )

    profile_distribution_dict = __build_simultaneous_profile_distribution(pp1, pp2)

    profile1_distribution = np.zeros(len(profile_distribution_dict))
    profile2_distribution = np.zeros(len(profile_distribution_dict))
    for idx, (v1, v2) in enumerate(profile_distribution_dict.values()):
        profile1_distribution[idx] = v1
        profile2_distribution[idx] = v2

    distribution_len = len(profile_distribution_dict)
    cost_matrix = np.zeros((distribution_len, distribution_len))

    n_candidates = len(pp1.candidates)
    all_ranking_tuples = list(profile_distribution_dict.keys())
    for idx1, rank1 in enumerate(all_ranking_tuples):
        for idx2 in range(idx1 + 1, len(all_ranking_tuples)):
            rank2 = all_ranking_tuples[idx2]
            cost_matrix[idx1, idx2] = compute_ranking_distance_on_ballot_graph(
                rank1, rank2, n_candidates
            )
            cost_matrix[idx2, idx1] = cost_matrix[idx1, idx2]

    return emd_via_scipy_linear_program(
        source_distribution=profile1_distribution,
        target_distribution=profile2_distribution,
        cost_matrix=cost_matrix,
    )


def lp_dist(
    pp1: RankProfile,
    pp2: RankProfile,
    p_value: Optional[Union[int, str]] = 1,
) -> int:
    r"""
    Computes the :math:`L_p` distance between two profiles.
    Use 'inf' for infinity norm.
    Assumes both elections share the same candidates.

    Args:
        pp1 (RankProfile): RankProfile for first profile.
        pp2 (RankProfile): RankProfile for second profile.
        p_value (Union[int, str], optional): :math:`L_p` distance parameter. Use "inf" for
            :math:`\infty`.

    Returns:
        int: :math:`L_p` distance between two profiles.
    """
    pp_list = [pp1, pp2]
    pp_2arry = profiles_to_ndarrys(pp_list)
    electA = pp_2arry[:, 0]
    electB = pp_2arry[:, 1]

    if isinstance(p_value, int):
        sum = 0
        for i in range(len(electA)):
            diff = (abs(electA[i] - electB[i])) ** p_value
            sum += diff
        lp_dist = sum ** (1 / p_value)
        return lp_dist

    elif p_value == "inf":
        diff = [abs(x - y) for x, y in zip(electA, electB)]
        return max(diff)

    else:
        raise ValueError("Unsupported input type")


def profiles_to_ndarrys(profiles: list[RankProfile]):
    """
    Converts a list of preference profiles into an ndarray. The columns
    represent each profile, rows are the cast ballots, and each element represents
    the frequency a ballot type occurs for a ``RankProfile``. Each column will sum to one
    since weights are standardized. This is useful for computing election :math:`L_p` distances
    between profiles.

    Args:
        profiles (list[RankProfile]): A list of PreferenceProfiles.

    Returns:
        numpy.ndarray: computed matrix of ballot frequencies.
    """
    cast_ballots: list = []
    profile_dicts: list[dict] = []
    for pp in profiles:
        election_dict = rank_profile_to_ranking_dict(pp, standardize=True)
        profile_dicts.append(election_dict)
        for key in election_dict.keys():
            if key not in cast_ballots:
                cast_ballots.append(key)
    combined_dict = {ranking: 0 for ranking in cast_ballots}
    rows = len(cast_ballots)
    cols = len(profile_dicts)
    electn_ndarry = np.zeros((rows, cols))

    for i in range(len(profile_dicts)):
        election = combined_dict | profile_dicts[i]
        elect_distr = [float(election[key]) for key in sorted(election.keys())]
        electn_ndarry[:, i] = elect_distr
    return electn_ndarry


def euclidean_dist(point1: np.ndarray, point2: np.ndarray) -> float:
    return float(np.linalg.norm(point1 - point2, ord=2))



from .abstract_ranking import RankingElection

from .plurality import (
    Plurality,
    SNTV,
)

from .borda import Borda
from .stv import FastSTV, STV, IRV, SequentialRCV
from .alaska import Alaska
from .top_two import TopTwo
from .dominating_sets import DominatingSets
from .condo_borda import CondoBorda
from .plurality_veto import PluralityVeto
from .random_dictator import RandomDictator
from .boosted_random_dictator import BoostedRandomDictator
from .ranked_pairs import RankedPairs


__all__ = [
    "RankingElection",
    "Plurality",
    "SNTV",
    "Borda",
    "STV",
    "FastSTV",
    "IRV",
    "SequentialRCV",
    "Alaska",
    "DominatingSets",
    "CondoBorda",
    "TopTwo",
    "PluralityVeto",
    "RandomDictator",
    "BoostedRandomDictator",
    "RankedPairs",
]



from votekit.ballot import ScoreBallot
from typing import Union


def remove_cand_score_ballot(
    removed: Union[str, list],
    ballot: ScoreBallot,
) -> ScoreBallot:
    """
    Removes specified candidate(s) from ballot.

    Args:
        removed (Union[str, list]): Candidate or list of candidates to be removed.
        ballot (ScoreBallot): Ballot to remove candidates from.

    Returns:
        ScoreBallot: Ballot with candidate(s) removed.
    """
    if isinstance(removed, str):
        removed = [removed]

    scores = (
        {c: s for c, s in ballot.scores.items() if c not in removed}
        if ballot.scores is not None
        else None
    )

    new_ballot = ScoreBallot(
        scores=scores if scores != dict() else None,
        weight=ballot.weight,
        voter_set=ballot.voter_set,
    )

    return new_ballot



from votekit.elections.election_types.ranking.abstract_ranking import RankingElection
from votekit.pref_profile import RankProfile
from votekit.elections.election_state import ElectionState
from votekit.utils import first_place_votes, elect_cands_from_set_ranking
from votekit.cleaning import remove_and_condense_rank_profile
from typing import Optional, Literal
from functools import partial


class Plurality(RankingElection):
    """
    Plurality election. Winners are the m candidates with the most first-place votes.

    Args:
        profile (RankProfile): Profile to conduct election on.
        m (int, optional): Number of seats to elect. Defaults to 1.
        tiebreak (str, optional): Tiebreak method to use. Options are None, 'random', and 'borda'.
            Defaults to None, in which case a tie raises a ValueError.
        fpv_tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.
            Only used by ``score_function`` parameter.

    """

    def __init__(
        self,
        profile: RankProfile,
        m: int = 1,
        tiebreak: Optional[str] = None,
        fpv_tie_convention: Literal["high", "low", "average"] = "average",
    ):
        if m <= 0:
            raise ValueError("m must be strictly positive")
        if len(profile.candidates_cast) < m:
            raise ValueError("Not enough candidates received votes to be elected.")
        self.m = m
        self.tiebreak = tiebreak
        super().__init__(
            profile,
            score_function=partial(
                first_place_votes, tie_convention=fpv_tie_convention
            ),
            sort_high_low=True,
        )

    def _is_finished(self):
        # single round election
        elected_cands = [c for s in self.get_elected() for c in s]

        if len(elected_cands) == self.m:
            return True
        return False

    def _run_step(
        self, profile: RankProfile, prev_state: ElectionState, store_states=False
    ) -> RankProfile:
        """
        Run one step of an election from the given profile and previous state.
        In a Plurality election, find the :math:`m` candidates with the highest first-place vote
        totals.

        Args:
            profile (RankProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            RankProfile: The profile of ballots after the round is completed.
        """
        # since fpv is the score function, the remaining cands from round 0
        # are ranked by fpv
        # raises a ValueError is tiebreak is None and a tie occurs.
        elected, remaining, tie_resolution = elect_cands_from_set_ranking(
            prev_state.remaining, self.m, profile=profile, tiebreak=self.tiebreak
        )

        new_profile = remove_and_condense_rank_profile(
            [c for s in elected for c in s], profile
        )

        if store_states:
            if self.score_function:
                scores = self.score_function(new_profile)
            else:
                raise ValueError()

            if tie_resolution:
                tiebreaks = {tie_resolution[0]: tie_resolution[1]}
            else:
                tiebreaks = {}

            new_state = ElectionState(
                round_number=1,  # single shot election
                remaining=remaining,
                elected=elected,
                scores=scores,
                tiebreaks=tiebreaks,
            )

            self.election_states.append(new_state)

        return new_profile


class SNTV(Plurality):
    """
    Wrapper for Plurality election. Winners are the m candidates with the most first-place votes.

    Args:
        profile (RankProfile): Profile to conduct election on.
        m (int, optional): Number of seats to elect. Defaults to 1.
        tiebreak (str, optional): Tiebreak method to use. Options are None, 'random', and 'borda'.
            Defaults to None, in which case a tie raises a ValueError.

    """

    def __init__(
        self, profile: RankProfile, m: int = 1, tiebreak: Optional[str] = None
    ):
        super().__init__(profile, m, tiebreak)



"""
Generate ranked preference profiles using the Spacial models

The main API functions in this module are:

- `onedim_spacial_profile_generator`: Generates a single preference profile using a one-dimensional
    spacial model.
- `spacial_profile_and_positions_generator`: Generates a single preference profile using a
    multi-dimensional spacial model with voters and candidates distributed according to specified
    distributions.
- `clustered_spacial_profile_and_positions_generator`: Generates a single preference profile using a
    clustered multi-dimensional spacial model where voters are clustered around candidates.
"""

import numpy as np
from numpy.typing import NDArray
import pandas as pd
from typing import Optional, Tuple, Callable, Dict, Any, Sequence

from votekit.metrics import euclidean_dist
from votekit.pref_profile import RankProfile

# =================================================
# ================= API Functions =================
# =================================================


def onedim_spacial_profile_generator(
    candidates: Sequence[str],
    number_of_ballots: int,
) -> RankProfile:
    """
    Generates a ranked preference profile where voters and candidates
    are positioned on a one-dimensional line according to a normal
    distribution. Voter preferences are determined by their proximity
    to candidates on this line.

    Args:
        number_of_ballots (int): The number of ballots to generate.

    Returns:
        RankProfile: A ranked preference profile object.
    """
    n_candidates = len(candidates)

    candidate_position_dict = {c: np.random.normal(0, 1) for c in candidates}
    voter_positions = np.random.normal(0, 1, number_of_ballots)

    ballot_pool = np.full(
        (number_of_ballots, n_candidates), frozenset("~"), dtype=object
    )

    for i, vp in enumerate(voter_positions):
        distance_tuples = [
            (c, abs(v - vp)) for c, v, in candidate_position_dict.items()
        ]
        candidate_ranking = np.array(
            [frozenset({t[0]}) for t in sorted(distance_tuples, key=lambda x: x[1])]
        )
        ballot_pool[i] = candidate_ranking

    df = pd.DataFrame(ballot_pool)
    df.index.name = "Ballot Index"
    df.columns = [f"Ranking_{i + 1}" for i in range(n_candidates)]
    df["Weight"] = 1
    df["Voter Set"] = [frozenset()] * len(df)
    return RankProfile(
        candidates=candidates,
        df=df,
        max_ranking_length=n_candidates,
    )


def spacial_profile_and_positions_generator(
    number_of_ballots: int,
    candidates: list[str],
    voter_dist: Callable[..., np.ndarray] = np.random.uniform,
    voter_dist_kwargs: Optional[Dict[str, Any]] = None,
    candidate_dist: Callable[..., np.ndarray] = np.random.uniform,
    candidate_dist_kwargs: Optional[Dict[str, Any]] = None,
    distance: Callable[[np.ndarray, np.ndarray], float] = euclidean_dist,
) -> Tuple[RankProfile, dict[str, np.ndarray], np.ndarray]:
    """
    Samples a metric position for number_of_ballots voters from
    the voter distribution. Samples a metric position for each candidate
    from the input candidate distribution. With sampled
    positions, this method then creates a ranked RankProfile in which
    voter's preferences are consistent with their distances to the candidates
    in the metric space.

    Args:
        number_of_ballots (int): The number of ballots to generate.
        by_bloc (bool): Dummy variable from parent class.

    Returns:
        Tuple[RankProfile, dict[str, numpy.ndarray], numpy.ndarray]:
            A tuple containing the preference profile object,
            a dictionary with each candidate's position in the metric
            space, and a matrix where each row is a single voter's position
            in the metric space.
    """
    if voter_dist_kwargs is None:
        if voter_dist is np.random.uniform:
            voter_dist_kwargs = {"low": 0.0, "high": 1.0, "size": 2.0}
        else:
            voter_dist_kwargs = {}

    try:
        voter_dist(**voter_dist_kwargs)
    except TypeError:
        raise TypeError("Invalid kwargs for the voter distribution.")

    if candidate_dist_kwargs is None:
        if candidate_dist is np.random.uniform:
            candidate_dist_kwargs = {"low": 0.0, "high": 1.0, "size": 2.0}
        else:
            candidate_dist_kwargs = {}

    try:
        candidate_dist(**candidate_dist_kwargs)
    except TypeError:
        raise TypeError("Invalid kwargs for the candidate distribution.")

    try:
        v = voter_dist(**voter_dist_kwargs)
        c = candidate_dist(**candidate_dist_kwargs)
        distance(v, c)
    except TypeError:
        raise TypeError(
            "Distance function is invalid or incompatible "
            "with voter/candidate distributions."
        )

    candidate_position_dict = {
        c: candidate_dist(**candidate_dist_kwargs) for c in candidates
    }
    voter_positions = np.array(
        [voter_dist(**voter_dist_kwargs) for _ in range(number_of_ballots)]
    )

    ballot_pool = np.full((number_of_ballots, len(candidates)), frozenset("~"))

    for i in range(number_of_ballots):
        distance_tuples = [
            (c, distance(voter_positions[i], c_position))
            for c, c_position, in candidate_position_dict.items()
        ]
        candidate_ranking = np.array(
            [frozenset({t[0]}) for t in sorted(distance_tuples, key=lambda x: x[1])]
        )
        ballot_pool[i] = candidate_ranking

    n_candidates = len(candidates)
    df = pd.DataFrame(ballot_pool)
    df.index.name = "Ballot Index"
    df.columns = [f"Ranking_{i + 1}" for i in range(n_candidates)]
    df["Weight"] = 1
    df["Voter Set"] = [frozenset()] * len(df)
    return (
        RankProfile(
            candidates=candidates,
            df=df,
            max_ranking_length=n_candidates,
        ),
        candidate_position_dict,
        voter_positions,
    )


def clustered_spacial_profile_and_positions_generator(
    number_of_ballots: dict[str, int],
    candidates: list[str],
    voter_dist: Callable[..., np.ndarray] = np.random.normal,
    voter_dist_kwargs: Optional[Dict[str, Any]] = None,
    candidate_dist: Callable[..., np.ndarray] = np.random.uniform,
    candidate_dist_kwargs: Optional[Dict[str, Any]] = None,
    distance: Callable[[np.ndarray, np.ndarray], float] = euclidean_dist,
) -> Tuple[RankProfile, dict[str, np.ndarray], np.ndarray]:
    """
    Samples a metric position for each candidate
    from the input candidate distribution. For each candidate, then sample
    number_of_ballots[candidate] metric positions for voters
    which will be centered around the candidate.
    With sampled positions, this method then creates a ranked RankProfile in which
    voter's preferences are consistent with their distances to the candidates
    in the metric space.

    Args:
        number_of_ballots (dict[str, int]): The number of voters attributed
                    to each candidate {candidate string: # voters}.
        by_bloc (bool): Dummy variable from parent class.

    Returns:
        Tuple[RankProfile, dict[str, numpy.ndarray], numpy.ndarray]:
            A tuple containing the preference profile object,
            a dictionary with each candidate's position in the metric
            space, and a matrix where each row is a single voter's position
            in the metric space.
    """
    if voter_dist_kwargs is None:
        if voter_dist is np.random.normal:
            voter_dist_kwargs = {
                "loc": 0,
                "std": np.array(1.0),
                "size": np.array(2.0),
            }
        else:
            voter_dist_kwargs = {}

    if voter_dist.__name__ not in ["normal", "laplace", "logistic", "gumbel"]:
        raise ValueError("Input voter distribution not supported.")

    try:
        voter_dist_kwargs["loc"] = 0
        voter_dist(**voter_dist_kwargs)
    except TypeError:
        raise TypeError("Invalid kwargs for the voter distribution.")

    if candidate_dist_kwargs is None:
        if candidate_dist is np.random.uniform:
            candidate_dist_kwargs = {"low": 0.0, "high": 1.0, "size": 2.0}
        else:
            candidate_dist_kwargs = {}

    try:
        candidate_dist(**candidate_dist_kwargs)
    except TypeError:
        raise TypeError("Invalid kwargs for the candidate distribution.")

    try:
        v = voter_dist(**voter_dist_kwargs)
        c = candidate_dist(**candidate_dist_kwargs)
        distance(v, c)
    except TypeError:
        raise TypeError(
            "Distance function is invalid or incompatible "
            "with voter/candidate distributions."
        )

    candidate_position_dict: dict[str, NDArray] = {
        c: candidate_dist(**candidate_dist_kwargs) for c in candidates
    }

    n_voters = sum(number_of_ballots.values())
    voter_positions = [np.zeros(2) for _ in range(n_voters)]
    vidx = 0
    for c, c_position in candidate_position_dict.items():  # type: ignore
        for _ in range(number_of_ballots[c]):  # type: ignore
            voter_dist_kwargs["loc"] = c_position
            voter_positions[vidx] = voter_dist(**voter_dist_kwargs)
            vidx += 1

    n_candidates = len(candidates)
    ballot_pool = np.full((n_voters, n_candidates), frozenset("~"), dtype=object)
    for i in range(len(voter_positions)):
        v_position = voter_positions[i]
        distance_tuples = [
            (c, distance(v_position, c_position))
            for c, c_position, in candidate_position_dict.items()
        ]
        candidate_ranking = np.array(
            [frozenset({t[0]}) for t in sorted(distance_tuples, key=lambda x: x[1])]
        )
        ballot_pool[i] = candidate_ranking

    voter_positions_array = np.vstack(voter_positions)

    df = pd.DataFrame(ballot_pool)
    df.index.name = "Ballot Index"
    df.columns = [f"Ranking_{i + 1}" for i in range(n_candidates)]
    df["Weight"] = 1
    df["Voter Set"] = [frozenset()] * len(df)
    return (
        RankProfile(
            candidates=candidates,
            df=df,
            max_ranking_length=n_candidates,
        ),
        candidate_position_dict,
        voter_positions_array,
    )



from votekit.elections.election_types.scores import GeneralRating
from votekit.pref_profile import ScoreProfile
from typing import Optional


class Approval(GeneralRating):
    """
    Approval election. Standard approval voting lets voters choose any subset of candidates to
    approve.  Winners are the :math:`m` candidates who received the most approval votes.

    Args:
        profile (ScoreProfile): Profile to conduct election on.
        m (int, optional): Number of seats to elect. Defaults to 1.
        tiebreak (str, optional): Tiebreak method to use. Options are None and 'random'.
            Defaults to None, in which case a tie raises a ValueError.

    """

    def __init__(
        self, profile: ScoreProfile, m: int = 1, tiebreak: Optional[str] = None
    ):
        # limit one per candidate,  but no total budget limit
        super().__init__(profile, m=m, L=1, tiebreak=tiebreak)


class BlocPlurality(GeneralRating):
    """
    Like approval voting, but there is a user-specified limit of :math:`k` approvals per voter.
    Most commonly, this would be run with :math:`k=m`.

    Args:
        profile (ScoreProfile): Profile to conduct election on.
        m (int, optional): Number of seats to elect. Defaults to 1.
        k (int, optional): Total budget per voter. Defaults to None, which results in ``m``
            approvals per voter.
        tiebreak (str, optional): Tiebreak method to use. Options are None and 'random'.
            Defaults to None, in which case a tie raises a ValueError.

    """

    def __init__(
        self,
        profile: ScoreProfile,
        m: int = 1,
        k: Optional[int] = None,
        tiebreak: Optional[str] = None,
    ):
        if k is None or k == 0:
            k = m
        # limit one per candidate, total budget limit k
        super().__init__(profile, m=m, L=1, k=k, tiebreak=tiebreak)



from .comentions import comention, comention_above, comentions_matrix
from .boost import boost_prob, boost_matrix
from .candidate_distance import candidate_distance, candidate_distance_matrix


__all__ = [
    "comention",
    "comention_above",
    "comentions_matrix",
    "boost_prob",
    "boost_matrix",
    "candidate_distance",
    "candidate_distance_matrix",
]



from votekit.ballot_generator.bloc_slate_generator.name_bradley_terry import (
    name_bt_profile_generator,
    name_bt_profiles_by_bloc_generator,
    name_bt_profile_generator_using_mcmc,
    name_bt_profiles_by_bloc_generator_using_mcmc,
)
from votekit.ballot_generator.bloc_slate_generator.cambridge import (
    cambridge_profile_generator,
    cambridge_profiles_by_bloc_generator,
)
from votekit.ballot_generator.bloc_slate_generator.name_plackett_luce import (
    name_pl_profile_generator,
    name_pl_profiles_by_bloc_generator,
)
from votekit.ballot_generator.bloc_slate_generator.cumulative import (
    name_cumulative_profile_generator,
    name_cumulative_ballot_generator_by_bloc,
)

from votekit.ballot_generator.bloc_slate_generator.slate_bradley_terry import (
    slate_bt_profile_generator,
    slate_bt_profiles_by_bloc_generator,
    slate_bt_profile_generator_using_mcmc,
    slate_bt_profiles_by_bloc_generator_using_mcmc,
)
from votekit.ballot_generator.bloc_slate_generator.slate_plackett_luce import (
    slate_pl_profile_generator,
    slate_pl_profiles_by_bloc_generator,
)
from votekit.ballot_generator.bloc_slate_generator.model import (
    convert_cohesion_map_to_cohesion_df,
    convert_preference_map_to_preference_df,
    convert_bloc_proportion_map_to_series,
    BlocSlateConfig,
)

__all__ = [
    "slate_bt_profile_generator",
    "slate_bt_profiles_by_bloc_generator",
    "slate_bt_profile_generator_using_mcmc",
    "slate_bt_profiles_by_bloc_generator_using_mcmc",
    "name_bt_profile_generator",
    "name_bt_profiles_by_bloc_generator",
    "name_bt_profile_generator_using_mcmc",
    "name_bt_profiles_by_bloc_generator_using_mcmc",
    "name_pl_profile_generator",
    "name_pl_profiles_by_bloc_generator",
    "name_cumulative_profile_generator",
    "name_cumulative_ballot_generator_by_bloc",
    "cambridge_profile_generator",
    "cambridge_profiles_by_bloc_generator",
    "slate_pl_profile_generator",
    "slate_pl_profiles_by_bloc_generator",
    "slate_bt_profile_generator",
    "slate_bt_profiles_by_bloc_generator",
    "convert_cohesion_map_to_cohesion_df",
    "convert_preference_map_to_preference_df",
    "convert_bloc_proportion_map_to_series",
    "BlocSlateConfig",
]



"""
Generate ranked preference profiles using the name-Plackett-Luce model.

The main API functions in this module are:

- `name_pl_profile_generator`: Generates a single preference profile using the name-Plackett-Luce
    model.
- `name_pl_profiles_by_bloc_generator`: Generates preference profiles by bloc using the
    name-Plackett-Luce model.
"""

import numpy as np
from typing import Optional
import pandas as pd
import apportionment.methods as apportion

from votekit.ballot_generator.bloc_slate_generator.model import BlocSlateConfig
from votekit.pref_profile import RankProfile


# ===========================================================
# ================= Interior Work Functions =================
# ===========================================================


def _inner_name_plackett_luce(
    config: BlocSlateConfig,
    *,
    ballot_length: Optional[int] = None,
) -> dict[str, RankProfile]:
    """
    Inner function to generate preference profiles by bloc using the name-Plackett-Luce model.

    The Plackett-Luce model samples without replacement from each preference interval
    for each bloc of voters.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        ballot_length (Optional[int]): Number of ranking positions allowed per ballot. If None,
            this is set to the total number of candidates in the configuration. Defaults to None.

    Returns:
        dict[str, RankProfile]: Dictionary whose keys are bloc strings and values are
            ``RankProfile`` objects.
    """
    n_candidates = len(config.candidates)
    if ballot_length is None:
        ballot_length = n_candidates

    bloc_lst = config.blocs

    bloc_counts = apportion.compute(
        "huntington", list(config.bloc_proportions.values()), config.n_voters
    )
    if not isinstance(bloc_counts, list):
        if not isinstance(bloc_counts, int):
            raise TypeError(
                f"Unexpected type from apportionment got {type(bloc_counts)}"
            )

        bloc_counts = [bloc_counts]

    ballots_per_bloc = {bloc: bloc_counts[i] for i, bloc in enumerate(bloc_lst)}

    pp_by_bloc = {b: RankProfile() for b in bloc_lst}
    pref_interval_by_bloc_dict = config.get_combined_preference_intervals_by_bloc()

    for bloc in config.blocs:
        n_ballots = ballots_per_bloc[bloc]
        ballot_pool = np.full((n_ballots, ballot_length), frozenset("~"))
        non_zero_cands = list(pref_interval_by_bloc_dict[bloc].non_zero_cands)
        pref_interval_values = [
            pref_interval_by_bloc_dict[bloc].interval[c] for c in non_zero_cands
        ]
        zero_cands = list(pref_interval_by_bloc_dict[bloc].zero_cands)

        # if there aren't enough non-zero supported candidates,
        # include 0 support as ties
        number_to_sample = ballot_length
        number_tied = None

        if len(non_zero_cands) < number_to_sample:
            number_tied = number_to_sample - len(non_zero_cands)
            number_to_sample = len(non_zero_cands)

        for i in range(n_ballots):
            non_zero_ranking = list(
                np.random.choice(
                    non_zero_cands,
                    number_to_sample,
                    p=pref_interval_values,
                    replace=False,
                )
            )

            ranking = [frozenset({cand}) for cand in non_zero_ranking]

            if number_tied is not None:
                ranking.append(frozenset(zero_cands))

            ballot_pool[i] = np.array(ranking)

        df = pd.DataFrame(ballot_pool)
        df.index.name = "Ballot Index"
        df.columns = [f"Ranking_{i + 1}" for i in range(n_candidates)]
        df["Weight"] = 1
        df["Voter Set"] = [frozenset()] * len(df)
        pp = RankProfile(
            candidates=config.candidates,
            df=df,
            max_ranking_length=ballot_length,
        )
        pp_by_bloc[bloc] = pp

    return pp_by_bloc


# =================================================
# ================= API Functions =================
# =================================================

# NOTE: The 'ballot_length' parameter has been commented out
# since we don't have good analogues in any of the other Bloc-Slate
# generators. When we add this feature, we can just restore the functionality.


def name_pl_profile_generator(
    config: BlocSlateConfig,
    *,
    # ballot_length: Optional[int] = None,
    group_ballots: bool = True,
) -> RankProfile:
    """
    Generates a merged preference profile using the name-Plackett-Luce model.

    The Plackett-Luce model samples without replacement from each preference interval
    for each bloc of voters. These profiles are then merged into a single profile.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.

    Returns:
        RankProfile: Merged ``RankProfile`` object generated by the model.
    """
    config.is_valid(raise_errors=True)

    # pp_by_bloc = _inner_name_plackett_luce(config, ballot_length=ballot_length)
    pp_by_bloc = _inner_name_plackett_luce(config)

    pp = RankProfile(ballots=tuple())
    for profile in pp_by_bloc.values():
        pp += profile

    if group_ballots:
        pp = pp.group_ballots()

    return pp


def name_pl_profiles_by_bloc_generator(
    config: BlocSlateConfig,
    *,
    # ballot_length: Optional[int] = None,
    group_ballots: bool = True,
) -> dict[str, RankProfile]:
    """
    Generates a dictionary mapping bloc names to preference profiles by bloc using the
    name-Plackett-Luce model.

    The Plackett-Luce model samples without replacement from each preference interval
    for each bloc of voters.

    Args:
        config (BlocSlateConfig): Configuration object containing all necessary parameters for
            working with a bloc-slate ballot generator.
        group_ballots (bool): If True, group identical ballots in the returned profile and
            set the weight accordingly. Defaults to True.

    Returns:
        dict[str, RankProfile]: Dictionary whose keys are bloc strings and values are
            ``RankProfile`` objects generated by the model for each bloc.
    """
    config.is_valid(raise_errors=True)

    # pp_by_bloc = _inner_name_plackett_luce(config, ballot_length=ballot_length)
    pp_by_bloc = _inner_name_plackett_luce(config)

    if group_ballots:
        for bloc, profile in pp_by_bloc.items():
            pp_by_bloc[bloc] = profile.group_ballots()

    return pp_by_bloc



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .pref_profile import PreferenceProfile, RankProfile, ScoreProfile

from votekit.ballot import Ballot, RankBallot, ScoreBallot
from typing import Optional, Sequence
import pandas as pd
from functools import partial
import numpy as np


def _convert_ranking_cols_to_ranking(
    row: pd.Series, max_ranking_length: int
) -> Optional[tuple[frozenset, ...]]:
    """
    Convert the ranking cols to a ranking tuple in profile.df.

    Args:
        row (pd.Series): Row of a profile.df.
        max_ranking_length (int, optional): The maximum length of a ranking.

    Returns:
        Optional[tuple[frozenset, ...]]: Ranking of ballot.

    Raises:
        ValueError: NaN values can only trail on a ranking.

    """
    ranking_cols_idxs = [f"Ranking_{i+1}" for i in range(max_ranking_length)]

    if any(idx not in row.index for idx in ranking_cols_idxs):
        raise ValueError(f"Row has improper ranking columns: {row.index}.")

    if any(
        row[col_idx] == frozenset({"~"})
        and not all(row[idx] == frozenset({"~"}) for idx in ranking_cols_idxs[i:])
        for i, col_idx in enumerate(ranking_cols_idxs)
    ):
        raise ValueError(
            f"Row {row} has '~' between valid ranking positions. "
            "'~' values can only trail on a ranking."
        )

    ranking = [
        row[col_idx] for col_idx in ranking_cols_idxs if row[col_idx] != frozenset("~")
    ]

    return tuple(ranking) if len(ranking) > 0 else None


def convert_row_to_rank_ballot(
    row: pd.Series, max_ranking_length: int = 0
) -> RankBallot:
    """
    Convert a row of a properly formatted profile.df to a Ballot.

    Args:
        row (pd.Series): Row of a profile.df.
        max_ranking_length (int, optional): The maximum length of a ranking. Defaults to 0, which
            is used for ballots with no ranking.

    Returns:
        RankBallot: Ballot corresponding to the row of the df.
    """
    ranking = None
    if max_ranking_length > 0:
        ranking = _convert_ranking_cols_to_ranking(row, max_ranking_length)
    voter_set = row["Voter Set"]
    weight = row["Weight"]

    return RankBallot(
        ranking=ranking,
        weight=weight,
        voter_set=voter_set,
    )


def convert_row_to_score_ballot(
    row: pd.Series, candidates: tuple[str, ...]
) -> ScoreBallot:
    """
    Convert a row of a properly formatted profile.df to a Ballot.

    Args:
        row (pd.Series): Row of a profile.df.
        candidates (tuple[str,...]): The name of the candidates.

    Returns:
        ScoreBallot: Ballot corresponding to the row of the df.
    """
    scores = {c: row[c] for c in candidates if c in row and not pd.isna(row[c])}
    voter_set = row["Voter Set"]
    weight = row["Weight"]

    return ScoreBallot(
        scores=scores if scores != dict() else None,
        weight=weight,
        voter_set=voter_set,
    )


def _df_to_rank_ballot_tuple(
    df: pd.DataFrame, candidates: tuple[str, ...], max_ranking_length: int = 0
) -> tuple[RankBallot, ...]:
    """
    Convert a properly formatted profile.df into a list of ballots.

    Args:
        df (pd.DataFrame): A profile.df.
        candidates (tuple[str,...]): The candidates.
        max_ranking_length (int, optional): The maximum length of a ranking. Defaults to 0, which
            is used for ballots with no ranking.

    Returns:
        tuple[RankBallot]: The tuple of ballots.
    """
    if df.empty:
        return tuple()

    return tuple(
        df.apply(  # type: ignore[call-overload]
            partial(
                convert_row_to_rank_ballot,
                max_ranking_length=max_ranking_length,
            ),
            axis="columns",
        )
    )


def rank_profile_to_ballot_dict(
    rank_profile: RankProfile, standardize: bool = False
) -> dict[RankBallot, float]:
    """
    Converts profile to dictionary with keys = ballots and
    values = corresponding total weights.

    Args:
        rank_profile (RankProfile): Profile to convert.
        standardize (bool, optional): If True, divides the weight of each ballot by the total
            weight. Defaults to False.

    Returns:
        dict[Ballot, float]:
            A dictionary with ballots (keys) and corresponding total weights (values).
    """
    tot_weight = rank_profile.total_ballot_wt
    di: dict = {}
    for ballot in rank_profile.ballots:
        weightless_ballot = Ballot(
            ranking=ballot.ranking,
            voter_set=ballot.voter_set,
        )
        weight = ballot.weight
        if standardize:
            weight /= tot_weight

        if weightless_ballot not in di.keys():
            di[weightless_ballot] = weight
        else:
            di[weightless_ballot] += weight
    return di


def score_profile_to_ballot_dict(
    score_profile: ScoreProfile, standardize: bool = False
) -> dict[ScoreBallot, float]:
    """
    Converts profile to dictionary with keys = ballots and
    values = corresponding total weights.

    Args:
        score_profile (ScoreProfile): Profile to convert.
        standardize (bool, optional): If True, divides the weight of each ballot by the total
            weight. Defaults to False.

    Returns:
        dict[Ballot, float]:
            A dictionary with ballots (keys) and corresponding total weights (values).
    """
    tot_weight = score_profile.total_ballot_wt
    di: dict = {}
    for ballot in score_profile.ballots:
        weightless_ballot = Ballot(
            scores=ballot.scores,
            voter_set=ballot.voter_set,
        )
        weight = ballot.weight
        if standardize:
            weight /= tot_weight

        if weightless_ballot not in di.keys():
            di[weightless_ballot] = weight
        else:
            di[weightless_ballot] += weight
    return di


def rank_profile_to_ranking_dict(
    rank_profile: RankProfile, standardize: bool = False
) -> dict[tuple[frozenset[str], ...], float]:
    """
    Converts profile to dictionary with keys = rankings and
    values = corresponding total weights.

    Args:
        rank_profile (RankProfile): Profile to convert.
        standardize (bool, optional): If True, divides the weight of each ballot by the total
            weight. Defaults to False.

    Returns:
        dict[tuple[frozenset[str],...], float]:
            A dictionary with rankings (keys) and corresponding total weights (values).

    Raises:
        TypeError: Profile must be a RankProfile.
    """
    from .pref_profile import RankProfile

    if not isinstance(rank_profile, RankProfile):
        raise TypeError(("Profile must be a RankProfile."))
    tot_weight = rank_profile.total_ballot_wt
    di: dict = {}
    for ballot in rank_profile.ballots:
        ranking = ballot.ranking
        weight = ballot.weight
        if standardize:
            weight /= tot_weight
        di[ranking] = di.get(ranking, 0) + weight

    return di


def score_profile_to_scores_dict(
    score_profile: ScoreProfile, standardize: bool = False
) -> dict[tuple[str, float], float]:
    """
    Converts profile to dictionary with keys = scores and
    values = corresponding total weights.

    Args:
        score_profile (ScoreProfile): Profile to convert.
        standardize (bool, optional): If True, divides the weight of each ballot by the total
            weight. Defaults to False.

    Returns:
        dict[tuple[str, float], float]:
            A dictionary with scores (keys) and corresponding total weights (values).

    Raises:
        TypeError: Profile must be a ScoreProfile.
    """
    from .pref_profile import ScoreProfile

    if not isinstance(score_profile, ScoreProfile):
        raise TypeError(("Profile must be a ScoreProfile."))

    tot_weight = score_profile.total_ballot_wt
    di: dict = {}
    for ballot in score_profile.ballots:
        scores = tuple(ballot.scores.items()) if ballot.scores else None
        weight = ballot.weight
        if standardize:
            weight /= tot_weight

        di[scores] = di.get(scores, 0) + weight
    return di


def profile_df_head(
    profile: PreferenceProfile,
    n: int,
    sort_by_weight: Optional[bool] = True,
    percents: Optional[bool] = False,
    totals: Optional[bool] = False,
    n_decimals: int = 1,
) -> pd.DataFrame:
    """
    Returns a pd.DataFrame with the top-n ballots in profile.

    Args:
        n (int): Number of ballots to view.
        sort_by_weight (bool, optional): If True, rank ballot from most to least votes.
            If sorting by weight, index resets. Defaults to True.
        percents (bool, optional): If True, show voter share for a given ballot.
            Defaults to False.
        totals (bool, optional): If True, show total values for Percent and Weight.
            Defaults to False.
        n_decimals (int, optional): Number of decimals to round to. Defaults to 1.

    Returns:
        pandas.DataFrame: A dataframe with top-n ballots.

    Raises:
        ZeroDivisionError: Profile has 0 total ballot weight; cannot show percentages.
    """

    if sort_by_weight:
        df = profile.df.sort_values(by="Weight", ascending=False).head(n).copy()

    else:
        df = profile.df.head(n).copy()

    df_col_num = len(df.columns)
    if percents:
        if profile.total_ballot_wt == 0:
            raise ZeroDivisionError(
                "Profile has 0 total ballot weight; cannot show percentages."
            )
        df["Percent"] = df["Weight"] / float(profile.total_ballot_wt)

    if totals:
        total_row = [""] * (df_col_num - 1) + [df["Weight"].sum()]
        if percents:
            total_row += [df["Percent"].sum()]
        df.loc["Total"] = total_row

    if percents:
        df["Percent"] = df["Percent"].apply(lambda x: f"{float(x):.{n_decimals}%}")

    return df


def profile_df_tail(
    profile: PreferenceProfile,
    n: int,
    sort_by_weight: Optional[bool] = True,
    percents: Optional[bool] = False,
    totals: Optional[bool] = False,
    n_decimals: int = 1,
) -> pd.DataFrame:
    """
    Returns a pd.DataFrame with the bottom-n ballots in profile.

    Args:
        n (int): Number of ballots to view.
        sort_by_weight (bool, optional): If True, rank ballot from least to most votes.
            Defaults to True.
        percents (bool, optional): If True, show voter share for a given ballot.
            Defaults to False.
        totals (bool, optional): If True, show total values for Percent and Weight.
            Defaults to False.
        n_decimals (int, optional): Number of decimals to round to. Defaults to 1.

    Returns:
        pandas.DataFrame: A data frame with bottom-n ballots.

    Raises:
        ZeroDivisionError: Profile has 0 total ballot weight; cannot show percentages.
    """
    if sort_by_weight:
        df = profile.df.sort_values(by="Weight", ascending=False).tail(n).copy()
    else:
        df = profile.df.tail(n).copy()

    df_col_num = len(df.columns)
    if percents:
        if profile.total_ballot_wt == 0:
            raise ZeroDivisionError(
                "Profile has 0 total ballot weight; cannot show percentages."
            )
        df["Percent"] = df["Weight"] / float(profile.total_ballot_wt)

    if totals:
        total_row = [""] * (df_col_num - 1) + [df["Weight"].sum()]
        if percents:
            total_row += [df["Percent"].sum()]
        df.loc["Total"] = total_row

    if percents:
        df["Percent"] = df["Percent"].apply(lambda x: f"{float(x):.{n_decimals}%}")

    return df


def convert_rank_profile_to_score_profile_via_score_vector(
    rank_profile: RankProfile,
    score_vector: Sequence[float],
) -> ScoreProfile:
    """
    Convert a rank profile to a score profile using a score vector. Ballots must
    not contain ties. Score vector
    should be non-increasing and non-negative.

    Args:
        rank_profile (RankProfile): Rank profile to convert.
        score_vector (Sequence[float]): Score vector to use.

    Returns:
        ScoreProfile: Score profile.

    Raises:
        ValueError: Ballots must not contain ties.
        ValueError: Score vector must be non-increasing and non-negative.
    """
    # here to prevent circular import
    from votekit.utils import validate_score_vector
    from votekit.pref_profile import ScoreProfile

    validate_score_vector(score_vector)
    score_vector = list(score_vector)

    assert rank_profile.max_ranking_length is not None
    if len(score_vector) < rank_profile.max_ranking_length:
        score_vector += [0] * (rank_profile.max_ranking_length - len(score_vector))

    ranking_cols = [
        f"Ranking_{i}" for i in range(1, rank_profile.max_ranking_length + 1)
    ]
    rankings_arr = rank_profile.df[ranking_cols].to_numpy(dtype=object).ravel(order="K")
    if any(len(x) > 1 for x in rankings_arr):
        raise ValueError("Ballots must not contain ties.")

    cand_to_score_list = {
        c: [np.nan for _ in range(len(rank_profile.df))]
        for c in rank_profile.candidates
    }

    for df_tuple in rank_profile.df[ranking_cols].itertuples():
        ballot_idx, ranking = df_tuple[0], df_tuple[1:]
        for ranking_pos, cand_set in enumerate(ranking):
            if cand_set == frozenset({"~"}):
                continue
            cand = next(iter(cand_set))  # no ties so this is unique
            cand_to_score_list[cand][ballot_idx] = (
                score_vector[ranking_pos] if score_vector[ranking_pos] > 0 else np.nan
            )

    new_df = pd.DataFrame(cand_to_score_list)
    new_df.index.name = "Ballot Index"
    new_df["Voter Set"] = rank_profile.df["Voter Set"]
    new_df["Weight"] = rank_profile.df["Weight"]

    return ScoreProfile(
        df=new_df,
        candidates=rank_profile.candidates,
    )



from .rank_csv_utils import (
    _validate_rank_csv_format,
    _parse_profile_data_from_rank_csv,
    _parse_ballot_from_rank_csv,
)

from .score_csv_utils import (
    _validate_score_csv_format,
    _parse_profile_data_from_score_csv,
    _parse_ballot_from_score_csv,
)

__all__ = [
    "_validate_rank_csv_format",
    "_parse_profile_data_from_rank_csv",
    "_parse_ballot_from_rank_csv",
    "_validate_score_csv_format",
    "_parse_profile_data_from_score_csv",
    "_parse_ballot_from_score_csv",
]



"""
Generate ranked preference profiles using the Impartial Culture (IC) model.

The main API functions in this module are:

- `ic_profile_generator`: Generates a single preference profile using the IC distribution.
"""

import math
import numpy as np
import random
from typing import Sequence, Optional
from functools import lru_cache
from collections import Counter

from votekit.pref_profile import RankProfile
from votekit.utils import index_to_lexicographic_ballot, build_df_from_ballot_samples

# ====================================================
# ================= Helper Functions =================
# ====================================================


@lru_cache
def _total_num_ballots(n_candidates: int, max_ballot_length: int) -> int:
    """
    Calculate the total number of valid ballots given n_candidates and max_ballot_length.

    Args:
        n_candidates (int): the number of candidates in the election
        max_ballot_length (int): the maximum length of a ballot

    Returns:
        int: the total number of valid ballots
    """
    return sum(
        math.comb(n_candidates, i) * math.factorial(i)
        for i in range(1, max_ballot_length + 1)
    )


# ===========================================================
# ================= Interior Work Functions =================
# ===========================================================


def _generate_profile_optimized_non_short(
    candidates: Sequence[str],
    number_of_ballots: int,
    max_ballot_length: Optional[int] = None,
) -> RankProfile:
    """
    Generate an IC preference profile using Fisher-Yates shuffle
    {number_of_ballots} times. Used to generate a profile when
    short ballots are disallowed

    Args:
        candidates (Sequence[str]): the list of candidates in the election
        number_of_ballots (int): the number of ballots to generate
        max_ballot_length (Optional[int]): the maximum length allowed in the profile. If None,
            defaults to the number of candidates. Defaults to None.

    Returns:
        RankProfile
    """
    num_cands = len(candidates)
    if max_ballot_length is None:
        max_ballot_length = num_cands
    ballots_as_ind = [
        tuple(np.random.choice(num_cands, size=max_ballot_length, replace=False))
        for _ in range(number_of_ballots)
    ]
    ballots_as_counter = Counter(ballots_as_ind)
    pp_df = build_df_from_ballot_samples(dict(ballots_as_counter), candidates)
    pp_df.index.name = "Ballot Index"
    return RankProfile(
        df=pp_df,
        max_ranking_length=len(candidates),
        candidates=candidates,
    )


def _generate_profile_optimized_with_short(
    candidates: Sequence[str],
    number_of_ballots: int,
    max_ballot_length: Optional[int] = None,
) -> RankProfile:
    """
    Generate an IC profile in the case where short ballots are
    allowed. Randomly sample indices between 0 and number_of_valid
    ballots, we do this {number_of_ballots} times. Then we convert
    the indices to ballots using a help function

    Args:
        candidates (Sequence[str]): the list of candidates in the election
        number_of_ballots (int): the number of ballots to generate for
            the profile
        max_ballot_length (Optional[int]): the maximum length allowed in the profile. If None,
            defaults to the number of candidates. Defaults to None.

    Returns:
        RankProfile
    """
    num_cands = len(candidates)
    if max_ballot_length is None:
        max_ballot_length = num_cands
    total_ballots = _total_num_ballots(num_cands, max_ballot_length)

    # sample indices (representing allowed ballots) uniformally at
    # random
    ballot_inds = [
        random.randint(0, total_ballots - 1) for _ in range(number_of_ballots)
    ]
    ballots_as_cand_ind = [
        tuple(
            index_to_lexicographic_ballot(
                ballot_ind, num_cands, max_ballot_length, _total_num_ballots
            )
        )
        for ballot_ind in ballot_inds
    ]

    # Instantiate the preference profile using a dataframe
    ballots_as_counter = Counter(ballots_as_cand_ind)
    pp_df = build_df_from_ballot_samples(dict(ballots_as_counter), candidates)
    pp_df.index.name = "Ballot Index"
    return RankProfile(
        df=pp_df,
        max_ranking_length=len(candidates),
        candidates=candidates,
    )


# =================================================
# ================= API Functions =================
# =================================================


def ic_profile_generator(
    candidates: Sequence[str],
    number_of_ballots: int,
    max_ballot_length: Optional[int] = None,
    allow_short_ballots: bool = False,
) -> RankProfile:
    """
    Impartial Culture model where each ballot is equally likely.
    Equivalent to the ballot simplex with an alpha value of infinity.

    Args:
        candidates (Sequence[str]): The list of candidates in the election.
        number_of_ballots (int): The number of ballots to generate for the profile.
        max_ballot_length (Optional[int]): Maximum length of each ballot. If None, defaults to
            the number of candidates.
        allow_short_ballots (bool, optional): Whether to allow short ballots.
            Defaults to False.

    Returns:
        RankProfile: The generated preference profile
    """
    if max_ballot_length is None:
        max_ballot_length = len(candidates)
    elif max_ballot_length > len(candidates):
        raise ValueError("Max ballot length larger than number of candidates given.")

    if allow_short_ballots:
        return _generate_profile_optimized_with_short(
            candidates, number_of_ballots, max_ballot_length
        )

    return _generate_profile_optimized_non_short(
        candidates, number_of_ballots, max_ballot_length
    )



from votekit.elections.election_types.ranking.abstract_ranking import RankingElection
from votekit.pref_profile import RankProfile
from votekit.elections.election_state import ElectionState
from votekit.cleaning import remove_and_condense_rank_profile
from votekit.graphs import PairwiseComparisonGraph


class DominatingSets(RankingElection):
    """
    A "dominating set" is any set S of candidates such that everyone in S beats everyone outside of
    S head-to-head. The top tier (which is well defined) is often called the "Smith set," and if it
    is just one person, they are called the "Condorcet candidate." The Smith method of election
    declares the Smith set to be the winners, which means that users do not get to specify the
    number of winners.

    Args:
        profile (RankProfile): Profile to conduct election on.

    """

    def __init__(self, profile: RankProfile):
        super().__init__(profile)

    def _is_finished(self):
        # single round election
        if len(self.election_states) == 2:
            return True
        return False

    def _run_step(
        self, profile: RankProfile, prev_state: ElectionState, store_states=False
    ) -> RankProfile:
        """
        Run one step of an election from the given profile and previous state.
        Computes the dominating tiers and returns the highest tier, candidates that beat
        every other candidate in head-to-head comparisons.

        Args:
            profile (RankProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            RankProfile: The profile of ballots after the round is completed.
        """

        pwc_graph = PairwiseComparisonGraph(profile)
        dominating_tiers = pwc_graph.get_dominating_tiers()
        new_profile = remove_and_condense_rank_profile(
            list(dominating_tiers[0]), profile
        )

        if store_states:
            elected = (frozenset(dominating_tiers[0]),)
            remaining = tuple([frozenset(s) for s in dominating_tiers[1:]])

            self.election_states.append(
                ElectionState(
                    round_number=prev_state.round_number + 1,
                    elected=elected,
                    remaining=remaining,
                )
            )

        return new_profile
